"""navigation.py contains the navigation items for the top level nav bar."""


from nautobot.core.apps import NavMenuAddButton, NavMenuGroup, NavMenuItem, NavMenuTab

menu_items = (
    NavMenuTab(
        name="Version Control",
        weight=1000,
        groups=(
            NavMenuGroup(
                name="Version Control",
                weight=100,
                items=(
                    NavMenuItem(
                        link="plugins:nautobot_version_control:branch_list",
                        name="Branches",
                        permissions=[],
                        buttons=(
                            NavMenuAddButton(
                                link="plugins:nautobot_version_control:branch_add",
                                permissions=[],
                            ),
                        ),
                    ),
                    NavMenuItem(
                        link="plugins:nautobot_version_control:pull_request_list",
                        name="Pull Requests",
                        permissions=[],
                        buttons=(
                            NavMenuAddButton(
                                link="plugins:nautobot_version_control:pull_request_add",
                                permissions=[],
                            ),
                        ),
                    ),
                    NavMenuItem(
                        link="plugins:nautobot_version_control:commit_list",
                        name="Commits (Active Branch)",
                        permissions=[],
                        buttons=(),
                    ),
                    NavMenuItem(
                        link="plugins:nautobot_version_control:active_branch_diffs",
                        name="Diffs (Active Branch)",
                        permissions=[],
                        buttons=(),
                    ),
                ),
            ),
        ),
    ),
)



# Generated by Django 3.1.13 on 2021-08-17 02:51

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("nautobot_version_control", "0005_pullrequest_pullrequestreview"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pullrequest",
            name="creator",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="users.user"),
        ),
        migrations.AlterField(
            model_name="pullrequest",
            name="description",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="pullrequestreview",
            name="reviewer",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="users.user"),
        ),
        migrations.AlterField(
            model_name="pullrequestreview",
            name="summary",
            field=models.TextField(blank=True, null=True),
        ),
    ]



""" This is the main module that contains the code for the Dolt backed Version Control plugin. """
try:
    from importlib import metadata
except ImportError:
    # Python version < 3.8
    import importlib_metadata as metadata

from django.db.models.signals import pre_migrate, post_migrate
import django_tables2

from nautobot.extras.plugins import PluginConfig

from nautobot_version_control.migrations import auto_dolt_commit_migration

__version__ = metadata.version(__name__)


class NautobotVersionControl(PluginConfig):
    """NautobotVersionControl initializes the dolt configs, middleware, and sets up migrations."""

    name = "nautobot_version_control"
    verbose_name = "Nautobot Version Control"
    description = "Nautobot Version Control with Dolt"
    base_url = "version-control"
    version = __version__
    author = "Network to Code, LLC"
    author_email = "opensource@networktocode.com"
    min_version = "1.2.0-beta.1"
    max_version = "1.999"
    required_settings = []
    default_settings = {
        # TODO: are these respected?
        #   this is also set in /development/nautobot_config.py
        "DATABASE_ROUTERS": [
            "nautobot_version_control.routers.GlobalStateRouter",
        ],
        "SESSION_ENGINE": "django.contrib.sessions.backends.signed_cookies",
        "CACHEOPS_ENABLED": False,
    }
    middleware = [
        "nautobot_version_control.middleware.dolt_health_check_intercept_middleware",
        "nautobot_version_control.middleware.DoltBranchMiddleware",
        "nautobot_version_control.middleware.DoltAutoCommitMiddleware",
    ]

    def ready(self):
        super().ready()

        # disable the GlobalStateRouter during migrations.
        pre_migrate.connect(switch_global_router_off, sender=self)
        post_migrate.connect(switch_global_router_on, sender=self)

        # make a Dolt commit to save database migrations.
        post_migrate.connect(auto_dolt_commit_migration, sender=self)


config = NautobotVersionControl  # pylint: disable=C0103


def query_registry(model, registry):
    """Performs a lookup on a content type registry.
    Args:
        model: a Django model class
        registry: a python dictionary like
            ```
            {
                "my_app_label": True,
                "my_other_model": {
                    "my_model": True,
                },
            }
            ```
            The type of `<value>` is specific to each
            registry. A return value of `None` signals
            that nothing is registered for that `model`.
    """
    app_label = model._meta.app_label
    model = model.__name__.lower()

    if app_label not in registry:
        return None
    if not isinstance(registry[app_label], dict):
        return registry[app_label]

    # subset specified
    if model not in registry[app_label]:
        return None
    return registry[app_label][model]


# Registry of Content Types of models that should be under version control.
# Top-level dict keys are app_labels. If the top-level dict value is `True`,
# then all models under that app_label are allowlisted.The top-level value
# may also be a nest dict containing a subset of version-controlled models
# within the app_label.
__VERSIONED_MODEL_REGISTRY___ = {
    "nautobot_version_control": {
        # Pull Requests are not versioned
        "pullrequest": False,
        "pullrequestreviewcomments": False,
        "pullrequestreviews": False,
        "branchmeta": False,
        "branch": False,
        # todo: calling the following "versioned" is odd.
        #   their contents are parameterized by branch
        #   changes, but they are not under VCS.
        "commit": True,
        "commitancestor": True,
        "conflicts": True,
        "constraintviolations": True,
    },
    "dcim": True,
    "circuits": True,
    "ipam": True,
    "virtualization": True,
    "taggit": True,
    "tenancy": True,
    "extras": {
        # TODO: what should be versioned from `extras`?
        "computedfield": True,
        "configcontext": True,
        "configcontextschema": True,
        "customfield": True,
        "customfieldchoice": True,
        "customlink": True,
        "exporttemplate": True,
        # "gitrepository": True,
        "graphqlquery": True,
        "imageattachment": True,
        # "job": True,
        # "jobresult": True,
        "objectchange": True,
        "relationship": True,
        "relationshipassociation": True,
        "secret": True,
        "secretsgroup": True,
        "status": True,
        "tag": True,
        "taggeditem": True,
        "webhook": True,
    },
}


def is_versioned_model(model):
    """
    Determines whether a model's is under version control.
    See __MODELS_UNDER_VERSION_CONTROL__ for more info.
    """
    registry = __VERSIONED_MODEL_REGISTRY___
    return bool(query_registry(model, registry))


def register_versioned_models(registry):
    """Register additional content types to be versioned.
    Args:
        registry: a python dict of content types that
            will be placed under version control:
            ```
            {
                "my_app_label": True,
                "my_other_model": {
                    "my_model": True,
                },
            }
            ```
    """
    err = ValueError("invalid versioned model registry")
    for key, val in registry.items():
        if not isinstance(key, str):
            # key must be string
            raise err
        if isinstance(val, bool):
            # val may be bool
            continue
        if not isinstance(val, dict):
            # val must be dict if not bool
            raise err
        # validate nested dict
        for k, v in val.items():
            if not isinstance(k, str):
                # k must be string
                raise err
            if not isinstance(v, bool):
                # v must be bool
                raise err
    __VERSIONED_MODEL_REGISTRY___.update(registry)


__DIFF_TABLE_REGISTRY__ = {}


def diff_table_for_model(model):
    """
    Returns a table object for a model, if it exists in
    the ` __DIFF_TABLE_REGISTRY__`.
    """
    return query_registry(model, __DIFF_TABLE_REGISTRY__)


def register_diff_tables(registry):
    """Register additional tables to be used in diffs.
    Registry values must be subclasses of django_tables2.Table.

    Args:
        registry: a python dict of content types that
            will be placed under version control:
            ```
            {
                "my_app_label": True,
                "my_other_model": {
                    "my_model": True,
                },
            }
            ```
    """
    err = ValueError("invalid diff table registry")
    for key, val in registry.items():
        if not isinstance(key, str):
            # key must be string
            raise err
        if not isinstance(val, dict):
            # val must be dict
            raise err
        for k, v in val.items():
            if not isinstance(k, str):
                # k must be string
                raise err
            if not issubclass(v, django_tables2.tables.Table):
                # v must be Table
                raise err
    __DIFF_TABLE_REGISTRY__.update(registry)


__GLOBAL_ROUTER_SWITCH__ = True


def is_global_router_enabled():
    """Returns true if the __GLOBAL_ROUTER_SWITCH__ is turned on"""
    global __GLOBAL_ROUTER_SWITCH__  # pylint: disable=W0602
    return __GLOBAL_ROUTER_SWITCH__


def switch_global_router_on(**kwargs):
    """Sets __GLOBAL_ROUTER_SWITCH to true"""

    global __GLOBAL_ROUTER_SWITCH__
    __GLOBAL_ROUTER_SWITCH__ = True


def switch_global_router_off(**kwargs):
    """Sets __GLOBAL_ROUTER_SWITCH to false"""
    global __GLOBAL_ROUTER_SWITCH__
    __GLOBAL_ROUTER_SWITCH__ = False



from django.db import connection


def auto_dolt_commit_migration(sender, **kwargs):
    msg = "Completed database migration"
    author = "system <nautobot@nautobot.invalid>"
    with connection.cursor() as cursor:
        cursor.execute("SELECT dolt_add('-A') FROM dual;")
        cursor.execute(
            f"""
            SELECT dolt_commit(
                '--all',
                '--allow-empty',
                '--message', '{msg}',
                '--author', '{author}')
            FROM dual;"""
        )



# Generated by Django 3.1.13 on 2021-07-30 15:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nautobot_version_control", "0003_conflicts"),
    ]

    operations = [
        migrations.CreateModel(
            name="ConstraintViolations",
            fields=[
                ("table", models.TextField(primary_key=True, serialize=False)),
                ("num_violations", models.IntegerField()),
            ],
            options={
                "verbose_name_plural": "constraint violations",
                "db_table": "dolt_constraint_violations",
                "managed": False,
            },
        ),
    ]



"""views.py implements django views for all Version Control plugin features."""

from datetime import datetime
import logging

from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404, get_list_or_404, render, redirect
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views import View

from nautobot.core.views import generic
from nautobot.dcim.models.sites import Site
from nautobot.utilities.forms import ConfirmationForm
from nautobot.utilities.permissions import get_permission_for_model
from nautobot.utilities.views import GetReturnURLMixin, ObjectPermissionRequiredMixin

from nautobot_version_control import diffs, filters, forms, merge, tables
from nautobot_version_control.constants import DOLT_DEFAULT_BRANCH
from nautobot_version_control.utils import alter_session_branch, db_for_commit, active_branch
from nautobot_version_control.models import (
    Branch,
    BranchMeta,
    Commit,
    CommitAncestor,
    PullRequest,
    PullRequestReview,
)


#
# Branches
#


class BranchView(generic.ObjectView):
    """BranchView renders a view of a Branch object."""

    queryset = Branch.objects.all()

    def get_extra_context(self, request, instance):  # pylint: disable=W0613,C0116 # noqa: D102
        merge_base = Commit.merge_base(DOLT_DEFAULT_BRANCH, instance.name)
        head = instance.hash
        return {"results": diffs.two_dot_diffs(from_commit=merge_base, to_commit=head)}


class BranchListView(generic.ObjectListView):
    """BranchListView renders a view of all branches."""

    queryset = Branch.objects.exclude(name__startswith="xxx")
    filterset = filters.BranchFilterSet
    filterset_form = forms.BranchFilterForm
    table = tables.BranchTable
    action_buttons = ("add",)
    template_name = "nautobot_version_control/branch_list.html"

    def extra_context(self):  # pylint: disable=W0613,C0116  # noqa: D102
        return {"default_branch": DOLT_DEFAULT_BRANCH}


class BranchCheckoutView(View):
    """BranchCheckoutView renders a view of checking out a branch."""

    queryset = Branch.objects.all()
    model_form = forms.BranchForm
    template_name = "nautobot_version_control/branch_edit.html"

    def get(self, req, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        # new branch will be checked out on redirect
        alter_session_branch(sess=req.session, branch=kwargs["pk"])
        return redirect("/")


class BranchEditView(generic.ObjectEditView):
    """BranchEditView renders either an add or create a branch."""

    queryset = Branch.objects.all()
    model_form = forms.BranchForm
    template_name = "nautobot_version_control/branch_edit.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        initial = {
            "starting_branch": Branch.objects.get(name=DOLT_DEFAULT_BRANCH),
            "creator": request.user,
        }
        return render(
            request,
            self.template_name,
            {
                "obj_type": self.queryset.model._meta.verbose_name,
                "form": self.model_form(initial=initial),
            },
        )

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116  # noqa: D102
        form = self.model_form(data=request.POST, files=request.FILES)
        response = super().post(request, *args, **kwargs)
        if BranchEditView._is_success_response(response):
            self.create_branch_meta(request, form)
            alter_session_branch(sess=request.session, branch=form.data.get("name"))
        return response

    @staticmethod
    def _is_success_response(response):
        """returns if response was successful."""
        return response.status_code // 100 in (2, 3)

    @staticmethod
    def create_branch_meta(req, form):
        """creates a BranchMeta object for a Branch."""
        meta, _ = BranchMeta.objects.get_or_create(branch=form.data.get("name"))
        meta.source_branch = form.data.get("starting_branch")
        meta.author = req.user
        meta.created = datetime.now()
        meta.save()


class BranchBulkEditView(generic.BulkEditView):
    """BranchBulkEditView is used to edit a set of branches at once."""

    queryset = Branch.objects.all()
    filterset = filters.BranchFilterSet
    table = tables.BranchTable
    form = forms.BranchBulkEditForm


class BranchBulkDeleteView(generic.BulkDeleteView):
    """BranchBulkDeleteView is used to delete a set of branches at once."""

    queryset = Branch.objects.all()
    table = tables.BranchTable
    form = forms.BranchBulkDeleteForm
    template_name = "nautobot_version_control/branch_bulk_delete.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        return redirect(self.get_return_url(request))

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116  # noqa: D102
        logger = logging.getLogger("nautobot.views.BulkDeleteView")
        model = self.queryset.model

        # Are we deleting *all* objects in the queryset or just a selected subset?
        if request.POST.get("_all"):
            pk_list = model.objects.values_list("pk", flat=True)
        else:
            pk_list = request.POST.getlist("pk")

        form_cls = self.get_form()

        if "_confirm" in request.POST:
            form = form_cls(request.POST)
            if form.is_valid():
                logger.debug("Form validation was successful")

                # Delete objects
                queryset = self.queryset.filter(pk__in=pk_list)
                try:
                    deleted_count = queryset.delete()[1][model._meta.label]
                except Exception as e:
                    logger.info("Caught error while attempting to delete objects")
                    messages.error(request, mark_safe(e))
                    return redirect(self.get_return_url(request))

                msg = f"Deleted {deleted_count} {model._meta.verbose_name_plural}"
                logger.info(msg)
                messages.success(request, msg)
                return redirect(self.get_return_url(request))

            logger.debug("Form validation failed")

        else:
            form = form_cls(
                initial={
                    "pk": pk_list,
                    "return_url": self.get_return_url(request),
                }
            )

        # Retrieve objects being deleted
        table = self.table(self.queryset.filter(pk__in=pk_list), orderable=False)
        if not table.rows:
            messages.warning(request, f"No {model._meta.verbose_name_plural} were selected for deletion.")
            return redirect(self.get_return_url(request))

        context = {
            "form": form,
            "obj_type_plural": model._meta.verbose_name_plural,
            "table": table,
            "return_url": self.get_return_url(request),
        }
        context.update(self.extra_context())
        return render(request, self.template_name, context)


#
#   Merge
#


class BranchMergeFormView(GetReturnURLMixin, View):
    """BranchMergeFormView is used to confirm a merge."""

    queryset = Branch.objects.all()
    form = forms.MergeForm
    template_name = "nautobot_version_control/branch_merge.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        initial = {
            # TODO: use branch meta source branch
            "destination_branch": Branch.objects.get(name=DOLT_DEFAULT_BRANCH),
            "source_branch": Branch.objects.get(name=kwargs["src"]),
        }
        return render(
            request,
            self.template_name,
            {
                "obj_type": self.queryset.model._meta.verbose_name,
                "form": self.form(initial=initial),
            },
        )

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116  # noqa: D102
        form = self.form(data=request.POST, files=request.FILES)
        if not form.is_valid():
            raise ValueError(form.errors)
        src = form.cleaned_data.get("source_branch")
        dest = form.cleaned_data.get("destination_branch")
        # todo: don't use redirect
        return redirect(
            reverse(
                "plugins:nautobot_version_control:branch_merge_preview",
                kwargs={"src": src, "dest": dest},
            )
        )


class BranchMergePreView(GetReturnURLMixin, View):
    """BranchMergePreView is used to render a preview of a merge."""

    queryset = Branch.objects.all()
    form = forms.MergePreviewForm
    template_name = "nautobot_version_control/branch_merge_preview.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        src = Branch.objects.get(name=kwargs["src"])
        dest = Branch.objects.get(name=kwargs["dest"])
        # render a disabled form with previously submitted data
        initial = {
            "source_branch": src,
            "destination_branch": Branch.objects.get(name=kwargs["dest"]),
        }
        return render(
            request,
            self.template_name,
            {
                "obj_type": self.queryset.model._meta.verbose_name,
                "form": self.form(initial=initial),
                **self.get_extra_context(request, src, dest),
            },
        )

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        src, dest = kwargs["src"], kwargs["dest"]
        Branch.objects.get(name=dest).merge(src, user=request.user)
        messages.info(request, mark_safe(f"<h4>merged branch <b>{src}</b> into <b>{dest}</b></h4>"))
        alter_session_branch(sess=request.session, branch=dest)
        return redirect("/")

    def get_extra_context(self, request, src, dest):  # pylint: disable=W0613,C0116,R0201 # noqa: D102
        merge_base_c = Commit.merge_base(src, dest)
        source_head = src.hash
        return {
            "results": diffs.two_dot_diffs(from_commit=merge_base_c, to_commit=source_head),
            "conflicts": merge.get_conflicts_for_merge(src, dest),
            "back_btn_url": reverse("plugins:nautobot_version_control:branch_merge", args=[src.name]),
        }


#
# Commits
#


class CommitView(generic.ObjectView):
    """CommitView is used to render a commit."""

    queryset = Commit.objects.all()

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        """
        Looks up the requested commit using a database revision
        to ensure the commit is accessible.
        todo: explain ancestor
        """
        anc = get_list_or_404(CommitAncestor.objects.all(), **kwargs)[0]
        db = db_for_commit(anc.commit_hash)
        instance = self.queryset.using(db).get(commit_hash=anc.commit_hash)

        if anc.parent_hash:
            diff = diffs.two_dot_diffs(from_commit=anc.parent_hash, to_commit=instance)
        else:
            # init commit has no parents
            diff = {}

        return render(
            request,
            self.get_template_name(),
            {
                "object": instance,
                "results": diff,
            },
        )


class CommitListView(generic.ObjectListView):
    """CommitListView is used to a render a list of Commits."""

    queryset = Commit.objects.all()
    filterset = filters.CommitFilterSet
    filterset_form = forms.CommitFilterForm
    table = tables.CommitTable
    template_name = "nautobot_version_control/commit_list.html"
    action_buttons = None

    def alter_queryset(self, request):  # noqa: D102
        if active_branch() != DOLT_DEFAULT_BRANCH:
            # only list commits on the current branch since the merge-base
            merge_base_hash = Commit.merge_base(DOLT_DEFAULT_BRANCH, active_branch())
            merge_base = Commit.objects.get(commit_hash=merge_base_hash)
            self.queryset = self.queryset.filter(date__gt=merge_base.date)
        return self.queryset

    def extra_context(self):  # pylint: disable=W0613,C0116 # noqa: D102
        return {"active_branch": active_branch()}


class CommitEditView(generic.ObjectEditView):
    """CommitEditView is used to edit a commit."""

    queryset = Commit.objects.all()
    model_form = forms.CommitForm
    template_name = "nautobot_version_control/commit_edit.html"


class CommitDeleteView(generic.ObjectDeleteView):
    """CommitDeleteView is used to render the deletion of a commit."""

    queryset = Commit.objects.all()


class CommitRevertView(GetReturnURLMixin, ObjectPermissionRequiredMixin, View):
    """Revert Commits in bulk."""

    queryset = Commit.objects.all()
    form = forms.CommitBulkRevertForm
    table = tables.CommitRevertTable
    template_name = "nautobot_version_control/commit_revert.html"

    def get_required_permission(self):  # noqa: D102
        return get_permission_for_model(self.queryset.model, "change")

    def get_return_url(self, req):  # pylint: disable=W0221,W0613 # noqa: D102
        return reverse("plugins:nautobot_version_control:commit_list")

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        model = self.queryset.model
        pk_list = request.POST.getlist("pk")
        commits = self.queryset.filter(commit_hash__in=pk_list)

        if commits.count() != len(pk_list):
            found = set([c.commit_hash for c in commits])  # pylint: disable=R1718
            missing = [f"<strong>{h}</strong>" for h in pk_list if h not in found]
            messages.warning(
                request,
                mark_safe(
                    f"""Cannot revert commit(s) {", ".join(missing)},
                        commits were not found on branch
                        <strong>{active_branch()}</strong>"""
                ),
            )
            return redirect(self.get_return_url(request))

        context = {
            "pk": pk_list,
            "active_branch": active_branch(),
            "obj_type_plural": model._meta.verbose_name_plural,
            "content_type": ContentType.objects.get_for_model(model),
            "return_url": self.get_return_url(request),
        }

        if "_revert" in request.POST:
            context.update(
                {
                    "form": self.form(initial={"pk": pk_list}),
                    "table": self.table(commits),
                }
            )
            return render(request, self.template_name, context)

        if "_confirm" in request.POST:
            form = self.form(request.POST)
            if form.is_valid():
                commits = form.cleaned_data["pk"]
                msgs = [f"""<strong>"{c.short_message}"</strong>""" for c in commits]
                try:
                    _ = Commit.revert(commits, request.user)
                except Exception as e:
                    # catch database error
                    messages.error(
                        request,
                        mark_safe(f"""Error reverting commits {", ".join(msgs)}: {e}"""),
                    )
                    return redirect(self.get_return_url(request))
                else:
                    messages.success(
                        request,
                        mark_safe(f"""Successfully reverted commits {", ".join(msgs)}"""),
                    )

        return redirect(self.get_return_url(request))


#
# Diffs
#


class ActiveBranchDiffs(View):
    """ActiveBranchDiffs is used to render the differences in the current active_branch."""

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        return redirect(
            reverse(
                "plugins:nautobot_version_control:branch",
                kwargs={
                    "pk": active_branch(),
                },
            )
        )


class DiffDetailView(View):
    """DiffDetailView is used to render complex diff between a from and to commit."""

    template_name = "nautobot_version_control/diff_detail.html"

    def get_required_permission(self):  # pylint: disable=R0201
        """returns permissions."""
        return get_permission_for_model(Site, "view")

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        self.model = self.get_model(kwargs)  # pylint: disable=W0201
        before_obj, after_obj = self.get_objs(kwargs)
        return render(
            request,
            self.template_name,
            {
                "title": self.title(before_obj, after_obj),
                "display_name": self.display_name(kwargs),
                "diff_obj": self.get_json_diff(before_obj, after_obj),
                **self.breadcrumb(kwargs),
            },
        )

    def get_model(self, kwargs):
        """get_model returns the underlying model."""
        return ContentType.objects.get(app_label=kwargs["app_label"], model=kwargs["model"]).model_class()

    @staticmethod
    def title(before_obj, after_obj):
        """title returns the title of a diff."""
        if before_obj and after_obj:
            return f"Updated {after_obj}"
        elif after_obj:
            return f"Added {after_obj}"
        else:
            return f"Deleted {before_obj}"

    def breadcrumb(self, kwargs):
        """Return a breadcrumb."""
        return {
            "breadcrumb": {
                "app_label": kwargs["app_label"],
                "model": self.display_name(kwargs),
                "from": self.match_commit(kwargs["from_commit"]),
                "to": self.match_commit(kwargs["to_commit"]),
            }
        }

    @staticmethod
    def match_commit(commit):
        """Replace `commit` with a more semantically meaningful identifier, if possible."""
        if Branch.objects.filter(hash=str(commit)).count() == 1:
            b = Branch.objects.get(hash=str(commit))
            url = b.get_absolute_url()
            return mark_safe(f"<a href='{url}'>{b}</a>")
        elif Commit.objects.filter(commit_hash=str(commit)).exists():
            c = Commit.objects.get(commit_hash=str(commit))
            url = c.get_absolute_url()
            return mark_safe(f"<a href='{url}'>{c}</a>")
        return commit

    def display_name(self, kwargs):
        """returns the verbose name of the model."""
        return self.get_model(kwargs)._meta.verbose_name.capitalize()

    def get_objs(self, kwargs):
        """Returns the commit objects for the before and after of a diff."""
        pk = kwargs["pk"]
        from_commit = kwargs["from_commit"]
        to_commit = kwargs["to_commit"]
        qs = self.model.objects.all()
        before_obj, after_obj = None, None

        from_qs = qs.using(db_for_commit(from_commit))
        if from_qs.filter(pk=pk).exists():
            before_obj = from_qs.get(pk=pk)
        to_qs = qs.using(db_for_commit(to_commit))
        if to_qs.filter(pk=pk).exists():
            after_obj = to_qs.get(pk=pk)
        return before_obj, after_obj

    def get_json_diff(self, before_obj, after_obj):
        """Returns the diff as json objs."""
        before_obj = self.serialize_obj(before_obj)
        after_obj = self.serialize_obj(after_obj)
        added = not before_obj
        removed = not after_obj

        json_diff = []
        for field in self.model.csv_headers:
            if field in before_obj or field in after_obj:
                before_val = before_obj.get(field, "")
                after_val = after_obj.get(field, "")

                before_style, after_style = "", ""
                if removed:
                    before_style = "bg-danger"
                elif added:
                    after_style = "bg-success"
                elif before_val != after_val:
                    before_style = "bg-danger"
                    after_style = "bg-success"

                json_diff.append(
                    {
                        "name": field,
                        "before_val": before_val,
                        "before_style": before_style,
                        "after_val": after_val,
                        "after_style": after_style,
                    }
                )

        return json_diff

    def serialize_obj(self, obj):
        """serialize_obj converts a model into a json object."""
        if not obj:
            return {}
        json_obj = {}
        fields = {f.name for f in obj._meta.fields}
        fields |= set(self.model.csv_headers)
        for field in fields:
            try:
                val = getattr(obj, field)
                val = str(val) if val else "-"
                json_obj[field] = val
            except AttributeError:
                continue
        return json_obj


#
# Pull Requests
#


class PullRequestListView(generic.ObjectListView):
    """PullRequestListView is used to render a lit of pull requests."""

    queryset = PullRequest.objects.all().order_by("-created_at")
    filterset = filters.PullRequestDefaultOpenFilterSet
    filterset_form = forms.PullRequestFilterForm
    table = tables.PullRequestTable
    action_buttons = ()
    template_name = "nautobot_version_control/pull_request_list.html"


class PullRequestBase(generic.ObjectView):
    """PullRequestBase contains the base information about a PullRequest."""

    queryset = PullRequest.objects.all()
    actions = ()

    def get_extra_context(self, request, obj, **kwargs):  # pylint: disable=W0613,C0116,W0237 # noqa: D102
        src, dest = obj.get_src_dest_branches()
        return {
            "counts": {
                "num_conflicts": merge.get_conflicts_count_for_merge(src, dest),
                "num_reviews": obj.num_reviews,
                "num_commits": obj.num_commits,
            }
        }


class PullRequestDiffView(PullRequestBase):
    """PullRequestDiffView displays a diff for a pull request."""

    template_name = "nautobot_version_control/pull_request/diffs.html"

    def get_extra_context(self, request, obj, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        ctx = super().get_extra_context(request, obj, **kwargs)
        head = Branch.objects.get(name=obj.source_branch).hash
        merge_base = Commit.merge_base(obj.source_branch, obj.destination_branch)
        ctx.update(
            {
                "active_tab": "diffs",
                "results": diffs.two_dot_diffs(from_commit=merge_base, to_commit=head),
            }
        )
        return ctx


class PullRequestConflictView(PullRequestBase):
    """PullRequestConflictView renders a conflict for a pull request."""

    template_name = "nautobot_version_control/pull_request/conflicts.html"

    def get_extra_context(self, request, obj, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        ctx = super().get_extra_context(request, obj, **kwargs)
        src = Branch.objects.get(name=obj.source_branch)
        dest = Branch.objects.get(name=obj.destination_branch)
        ctx.update(
            {
                "active_tab": "conflicts",
                "conflicts": merge.get_conflicts_for_merge(src, dest),
            }
        )
        return ctx


class PullRequestReviewListView(PullRequestBase):
    """PullRequestReviewListView renders a list of pull requests."""

    template_name = "nautobot_version_control/pull_request/review_list.html"

    def get_extra_context(self, request, obj, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        ctx = super().get_extra_context(request, obj, **kwargs)
        reviews = PullRequestReview.objects.filter(pull_request=obj.pk).order_by("reviewed_at")
        ctx.update(
            {
                "active_tab": "reviews",
                "review_list": reviews,
            }
        )
        return ctx


class PullRequestCommitListView(PullRequestBase):
    """PullRequestCommitListView renders a list of commits."""

    template_name = "nautobot_version_control/pull_request/commits.html"
    table = tables.CommitTable

    def get_extra_context(self, request, obj, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        ctx = super().get_extra_context(request, obj, **kwargs)
        ctx.update(
            {
                "active_tab": "commits",
                "commits_table": self.table(obj.commits),
            }
        )
        return ctx


class PullRequestEditView(generic.ObjectEditView):
    """PullRequestEditView renders an edit view for a PR."""

    queryset = PullRequest.objects.all()
    model_form = forms.PullRequestForm
    template_name = "nautobot_version_control/pull_request/edit.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        obj = self.get_object(kwargs)
        initial = {
            "destination_branch": Branch.objects.get(name=DOLT_DEFAULT_BRANCH),
            "source_branch": active_branch(),
        }

        if obj.present_in_database:
            initial["title"] = obj.title
            initial["source_branch"] = obj.source_branch
            initial["destination_branch"] = obj.destination_branch
            initial["description"] = obj.description

        # overwrite with any query strings
        query_string_source = request.GET.get("source_branch", "")
        if query_string_source != "":
            initial["source_branch"] = query_string_source

        query_string_destination = request.GET.get("destination_branch", "")
        if query_string_destination != "":
            initial["destination_branch"] = query_string_destination

        return render(
            request,
            self.template_name,
            {
                "obj": obj,
                "obj_type": self.queryset.model._meta.verbose_name,
                "form": self.model_form(initial=initial),
                "return_url": self.get_return_url(request, obj),
                "editing": obj.present_in_database,
            },
        )

    def alter_obj(self, obj, request, url_args, url_kwargs):  # noqa: D102
        # Allow views to add extra info to an object before it is processed. For example, a parent object can be defined
        # given some parameter from the request URL.
        obj.creator = request.user
        return obj


class PullRequestReviewEditView(generic.ObjectEditView):
    """PullRequestReviewEditView renders an edit view for a PullRequestReview."""

    queryset = PullRequestReview.objects.all()
    model_form = forms.PullRequestReviewForm
    template_name = "nautobot_version_control/pull_request/review_edit.html"

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        initial = {
            "pull_request": PullRequest.objects.get(pk=kwargs["pull_request"]),
        }
        return render(
            request,
            self.template_name,
            {
                "obj_type": self.queryset.model._meta.verbose_name,
                "form": self.model_form(initial=initial),
            },
        )

    def get_return_url(self, req, obj):  # pylint: disable=W0613,W0237 # noqa: D102
        return reverse(
            "plugins:nautobot_version_control:pull_request_reviews",
            kwargs={"pk": obj.pull_request.pk},
        )

    def alter_obj(self, obj, request, url_args, url_kwargs):  # noqa: D102
        obj.reviewer = request.user
        return obj

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        kwargs["user"] = request.user
        return super().post(request, *args, **kwargs)


class PullRequestMergeView(generic.ObjectEditView):
    """PullRequestMergeView renders a view for rendering a pull request review."""

    queryset = PullRequest.objects.all()
    form = ConfirmationForm()
    template_name = "nautobot_version_control/pull_request/confirm_merge.html"

    def get(self, request, pk):  # pylint: disable=W0613,C0116,W0221 # noqa: D102
        pr = get_object_or_404(self.queryset, pk=pk)
        if pr.state != PullRequest.OPEN:
            msg = mark_safe(f"""Pull request "{pr}" is not open and cannot be merged""")
            messages.error(request, msg)
            return redirect("plugins:nautobot_version_control:pull_request", pk=pr.pk)
        src = Branch.objects.get(name=pr.source_branch)
        dest = Branch.objects.get(name=pr.destination_branch)
        return render(
            request,
            self.template_name,
            {
                "pull_request": pr,
                "form": self.form,
                "return_url": pr.get_absolute_url(),
                "conflicts": merge.get_conflicts_for_merge(src, dest),
                "diffs": diffs.three_dot_diffs(from_commit=dest.hash, to_commit=src.hash),
            },
        )

    def post(self, request, pk):  # pylint: disable=W0613,C0116,W0221 # noqa: D102
        pr = get_object_or_404(self.queryset, pk=pk)
        form = ConfirmationForm(request.POST)
        squash_param = request.POST.get("merge_squash", False)
        if squash_param == "true":
            squash_param = True

        if form.is_valid():
            pr.merge(user=request.user, squash=squash_param)
            messages.success(
                request,
                mark_safe(f"""Pull Request <strong>"{pr}"</strong> has been merged."""),
            )
            return redirect("plugins:nautobot_version_control:pull_request", pk=pr.pk)

        return render(
            request,
            self.template_name,
            {
                "pull_request": pr,
                "form": self.form,
                "return_url": pr.get_absolute_url(),
            },
        )


class PullRequestCloseView(generic.ObjectEditView):
    """PullRequestCloseView renders a view for closing a pr."""

    queryset = PullRequest.objects.all()
    form = ConfirmationForm()
    template_name = "nautobot_version_control/pull_request/confirm_close.html"

    def get(self, request, pk):  # pylint: disable=W0613,C0116,W0221 # noqa: D102
        pr = get_object_or_404(self.queryset, pk=pk)
        if pr.state != PullRequest.OPEN:
            msg = mark_safe(f"""Pull request "{pr}" is not open and cannot be closed""")
            messages.error(request, msg)
            return redirect("plugins:nautobot_version_control:pull_request", pk=pr.pk)

        return render(
            request,
            self.template_name,
            {
                "pull_request": pr,
                "form": self.form,
                "panel_class": "default",
                "button_class": "primary",
                "return_url": pr.get_absolute_url(),
            },
        )

    def post(self, request, pk):  # pylint: disable=W0221 # noqa: D102
        pr = get_object_or_404(self.queryset, pk=pk)
        form = ConfirmationForm(request.POST)

        if form.is_valid():
            pr.state = PullRequest.CLOSED
            pr.save()
            msg = mark_safe(f"""<strong>Pull Request "{pr}" has been closed.</strong>""")
            messages.success(request, msg)
            return redirect("plugins:nautobot_version_control:pull_request", pk=pr.pk)

        return render(
            request,
            self.template_name,
            {
                "pull_request": pr,
                "form": self.form,
                "return_url": pr.get_absolute_url(),
            },
        )


class PullRequestBulkDeleteView(generic.BulkDeleteView):
    """PullRequestBulkDeleteView renders a bulk delete form for a list of pull requests."""

    queryset = PullRequest.objects.all()
    table = tables.PullRequestTable
    form = forms.PullRequestDeleteForm
    template_name = "nautobot_version_control/pull_request_bulk_delete.html"

    def get(self, request):  # pylint: disable=W0613,C0116 # noqa: D102
        return redirect(self.get_return_url(request))

    def post(self, request, *args, **kwargs):  # pylint: disable=W0613,C0116 # noqa: D102
        logger = logging.getLogger("nautobot.views.BulkDeleteView")
        model = self.queryset.model

        # Are we deleting *all* objects in the queryset or just a selected subset?
        pk_list = request.POST.getlist("pk")
        form_cls = self.get_form()

        if "_confirm" in request.POST:
            form = form_cls(request.POST)
            if form.is_valid():
                logger.debug("Form validation was successful")

                # Delete objects
                queryset = self.queryset.filter(pk__in=pk_list)
                try:
                    deleted_count = queryset.delete()[1][model._meta.label]
                except Exception:
                    logger.info("Caught error while attempting to delete objects")
                    return redirect(self.get_return_url(request))

                msg = f"Deleted {deleted_count} {model._meta.verbose_name_plural}"
                logger.info(msg)
                messages.success(request, msg)
                return redirect(self.get_return_url(request))

            else:
                logger.debug("Form validation failed")

        else:
            form = form_cls(
                initial={
                    "pk": pk_list,
                    "return_url": self.get_return_url(request),
                }
            )

        # Retrieve objects being deleted
        table = self.table(self.queryset.filter(pk__in=pk_list), orderable=False)
        if not table.rows:
            messages.warning(
                request,
                f"No {model._meta.verbose_name_plural} were selected for deletion.",
            )
            return redirect(self.get_return_url(request))

        context = {
            "form": form,
            "obj_type_plural": model._meta.verbose_name_plural,
            "table": table,
            "return_url": self.get_return_url(request),
        }
        context.update(self.extra_context())
        return render(request, self.template_name, context)



from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Branch",
            fields=[
                ("name", models.TextField(primary_key=True, serialize=False)),
                ("hash", models.TextField()),
                ("latest_committer", models.TextField()),
                ("latest_committer_email", models.TextField()),
                ("latest_commit_date", models.DateTimeField()),
                ("latest_commit_message", models.TextField()),
            ],
            options={
                "verbose_name_plural": "branches",
                "db_table": "dolt_branches",
                "managed": False,
            },
        ),
        migrations.CreateModel(
            name="Commit",
            fields=[
                ("commit_hash", models.TextField(primary_key=True, serialize=False)),
                ("committer", models.TextField()),
                ("email", models.TextField()),
                ("date", models.DateTimeField()),
                ("message", models.TextField()),
            ],
            options={
                "verbose_name_plural": "commits",
                "db_table": "dolt_log",
                "managed": False,
            },
        ),
        migrations.CreateModel(
            name="CommitAncestor",
            fields=[
                ("commit_hash", models.TextField(primary_key=True, serialize=False)),
                ("parent_hash", models.TextField()),
                ("parent_index", models.IntegerField()),
            ],
            options={
                "verbose_name_plural": "commit_ancestors",
                "db_table": "dolt_commit_ancestors",
                "managed": False,
            },
        ),
    ]



"""Forms.py defines the set of forms used for creating, deleting, editing, and searching different models."""

from django import forms

from nautobot.users.models import User
from nautobot.utilities.forms import BootstrapMixin, ConfirmationForm, add_blank_choice

from nautobot_version_control.models import Branch, Commit, PullRequest, PullRequestReview
from nautobot_version_control.utils import active_branch, DoltError
from nautobot_version_control.constants import DOLT_DEFAULT_BRANCH


#
# Branches
#


class BranchForm(forms.ModelForm, BootstrapMixin):
    """BranchForm returns a form for the BranchCheckout and BranchEdit views."""

    name = forms.SlugField()
    starting_branch = forms.ModelChoiceField(queryset=Branch.objects.all(), to_field_name="name", required=True)
    creator = forms.ModelChoiceField(
        queryset=User.objects.all(),
        to_field_name="username",
        required=True,
        widget=forms.HiddenInput(),
    )

    class Meta:
        model = Branch
        fields = [
            "name",
            "starting_branch",
        ]

    def __init__(self, *args, **kwargs):
        # self.user = kwargs.pop("user", None)
        super().__init__(*args, **kwargs)

    def save(self, *args, **kwargs):
        """save overrides an superclass method."""
        self.instance.starting_branch = self.cleaned_data["starting_branch"]
        self.instance.creator = self.cleaned_data["creator"]
        return super().save(*args, **kwargs)


class MergeForm(forms.Form, BootstrapMixin):
    """MergeForm returns a form the merge button on the branch_merge page."""

    source_branch = forms.ModelChoiceField(queryset=Branch.objects.all(), to_field_name="name", required=True)
    destination_branch = forms.ModelChoiceField(queryset=Branch.objects.all(), to_field_name="name", required=True)

    class Meta:
        fields = [
            "source_branch",
            "destination_branch",
        ]


class MergePreviewForm(forms.Form, BootstrapMixin):
    """MergePreviewForm returns a form for previewing branch merges."""

    source_branch = forms.ModelChoiceField(
        queryset=Branch.objects.all(),
        to_field_name="name",
        disabled=True,
    )
    destination_branch = forms.ModelChoiceField(
        queryset=Branch.objects.all(),
        to_field_name="name",
        disabled=True,
    )

    class Meta:
        fields = [
            "source_branch",
            "destination_branch",
        ]


class BranchBulkEditForm(forms.Form, BootstrapMixin):
    """BranchBulkEditForm is a small form for the branch bulk edit view."""

    class Meta:
        model = Branch
        fields = [
            "name",
        ]


class BranchBulkDeleteForm(ConfirmationForm):
    """
    BranchBulkDeleteForm is used for validating the deletion of branches. It has additional checks for preventing
    deletes of the active branch and the DOLT_DEFAULT_BRANCH.
    """

    pk = forms.ModelMultipleChoiceField(queryset=Branch.objects.all(), widget=forms.MultipleHiddenInput)

    def clean_pk(self):
        """clean_pk gets the primary key of the cleaned data and protects against active_branch and default_branch deletions."""
        # TODO: log error messages
        deletes = [str(b) for b in self.cleaned_data["pk"]]
        if active_branch() in deletes:
            raise DoltError(f"Cannot delete active branch: {active_branch()}")
        if DOLT_DEFAULT_BRANCH in deletes:
            raise DoltError(f"Cannot delete primary branch: {DOLT_DEFAULT_BRANCH}")
        return self.cleaned_data["pk"]

    class Meta:
        model = Branch
        fields = [
            "name",
        ]


class BranchFilterForm(forms.Form, BootstrapMixin):
    """BranchFilterForm is used for filtering the branches list page."""

    model = Branch
    field_order = ["q"]
    q = forms.CharField(required=False, label="Search")

    latest_committer = forms.ChoiceField(choices=[], required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["latest_committer"].choices = add_blank_choice(
            Branch.objects.all().values_list("latest_committer", "latest_committer").distinct()
        )


#
# Commits
#


class CommitForm(forms.ModelForm, BootstrapMixin):
    """CommitForm is a form used for the CommitEdit view."""

    class Meta:
        model = Commit
        fields = ["message"]


class CommitFilterForm(forms.Form, BootstrapMixin):
    """CommitFilterForm is used to filter the commit set."""

    model = Commit
    field_order = ["q"]
    q = forms.CharField(required=False, label="Search")
    committer = forms.ChoiceField(choices=[], required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["committer"].choices = Commit.objects.all().values_list("committer", "committer").distinct()


class CommitBulkRevertForm(forms.Form, BootstrapMixin):
    """CommitBulkRevertForm is used to confirm the deletion of a certain amount of Commits."""

    pk = forms.ModelMultipleChoiceField(queryset=Commit.objects.all(), widget=forms.MultipleHiddenInput())

    class Meta:
        fields = ["branch"]


#
# PullRequests
#


class PullRequestForm(forms.ModelForm, BootstrapMixin):
    """PullRequestForms is used to confirm the creation or addition of pull requests."""

    qs = Branch.objects.exclude(name__startswith="xxx")
    source_branch = forms.ModelChoiceField(queryset=qs, to_field_name="name", required=True)
    destination_branch = forms.ModelChoiceField(queryset=qs, to_field_name="name", required=True)

    class Meta:
        model = PullRequest
        fields = [
            "title",
            "source_branch",
            "destination_branch",
            "description",
        ]


class PullRequestDeleteForm(ConfirmationForm):
    """PullRequestDeleteForm is used to delete selection PRs."""

    pk = forms.ModelMultipleChoiceField(queryset=PullRequest.objects.all(), widget=forms.MultipleHiddenInput)

    def clean_pk(self):
        """clean_pk returns only the pk of the cleaned data"""
        return self.cleaned_data["pk"]

    class Meta:
        model = PullRequest
        fields = [
            "pk",
        ]


class PullRequestFilterForm(forms.Form, BootstrapMixin):
    """PullRequestFilterForm is used to filter the complete PullRequest filter list."""

    model = PullRequest
    q = forms.CharField(required=False, label="Search")
    state = forms.MultipleChoiceField(required=False, choices=PullRequest.PR_STATE_CHOICES)
    creator = forms.ModelChoiceField(required=False, queryset=User.objects.all())
    reviewer = forms.ModelChoiceField(required=False, queryset=User.objects.all())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if args is not None:
            if not args[0].get("state", None):
                new_args = args[0].copy()
                new_args.update({"state": PullRequest.OPEN})
                super().__init__(new_args, **kwargs)

    class Meta:
        fields = [
            "state",
            "title",
            "source_branch",
            "destination_branch",
            "description",
        ]


class PullRequestReviewForm(forms.ModelForm, BootstrapMixin):
    """PullRequestReviewForm is used to review a pull request."""

    class Meta:
        model = PullRequestReview
        fields = [
            "pull_request",
            "summary",
            "state",
        ]



"""routers.py manages the GlobalStateRouter."""

from django.utils.safestring import mark_safe
from nautobot_version_control.constants import DOLT_DEFAULT_BRANCH, GLOBAL_DB
from nautobot_version_control.utils import DoltError, is_dolt_model, active_branch

from . import is_global_router_enabled, is_versioned_model


class GlobalStateRouter:
    """GlobalStateRouter manages the correct db to write either branch specific state or global state."""

    global_db = GLOBAL_DB

    def db_for_read(self, model, **hints):  # pylint: disable=W0613
        """
        Directs read queries to the global state db for non-versioned models.
        Versioned models use the 'default' database and the Dolt branch that
        was checked out in `DoltBranchMiddleware`.
        """
        if not is_global_router_enabled():
            return None

        if is_versioned_model(model):
            return None

        return self.global_db

    def db_for_write(self, model, **hints):  # pylint: disable=W0613
        """
        Directs write queries to the global state db for non-versioned models.
        Versioned models use the 'default' database and the Dolt branch that
        was checked out in `DoltBranchMiddleware`.
        Prevents writes of non-versioned models on non-primary branches.
        """
        if not is_global_router_enabled():
            return None

        if is_dolt_model(model):
            # Dolt models can be created or edited from any branch.
            # Edits will be applied to the "main"
            return self.global_db

        if is_versioned_model(model):
            return None

        if self.branch_is_not_primary():
            # non-versioned models can only be edited on "main"
            raise DoltError(
                mark_safe(
                    f"""Error writing model <strong>{model.__name__}</strong>
                        on branch <strong>"{active_branch()}"</strong>:
                        non-versioned models must be written on branch
                        <strong>"{DOLT_DEFAULT_BRANCH}"</strong>."""
                )
            )

        return self.global_db

    def allow_relation(self, obj1, obj2, **hints):  # pylint: disable=W0613.R0201
        """allow_relation allows a relation between obj1 and obj2 too exist."""
        return True

    @staticmethod
    def branch_is_not_primary():
        """branch_is_not_primary returns whether the active_branch is the default branch."""
        return active_branch() != DOLT_DEFAULT_BRANCH



"""models.py exposes Dolt primitives such as branches and commits as Django models."""


from django.core.exceptions import ObjectDoesNotExist
from django.db import models, connection, connections
from django.db.models import Q
from django.db.models.deletion import CASCADE
from django.urls import reverse
from django.utils.html import mark_safe, format_html
from django.dispatch import receiver
from django.db.models.signals import pre_delete

from nautobot.core.models import BaseModel
from nautobot.extras.utils import extras_features
from nautobot.users.models import User
from nautobot.utilities.querysets import RestrictedQuerySet

from nautobot_version_control.utils import author_from_user, DoltError, db_for_commit, active_branch
from nautobot_version_control.constants import DOLT_DEFAULT_BRANCH


class DoltSystemTable(models.Model):
    """DoltSystemTable represents an abstraction over Dolt builtin system tables."""

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        """Meta links to a _meta table."""

        abstract = True
        managed = False

    def validated_save(self):
        """Perform model validation during instance save."""
        self.full_clean()
        self.save()


#
# Branches
#


class Branch(DoltSystemTable):
    """Branch represents a model over the dolt_branches system table."""

    name = models.TextField(primary_key=True)
    hash = models.TextField()
    latest_committer = models.TextField()
    latest_committer_email = models.TextField()
    latest_commit_date = models.DateTimeField()
    latest_commit_message = models.TextField()

    class Meta:
        """Meta class."""

        managed = False
        db_table = "dolt_branches"
        verbose_name_plural = "branches"

    def __init__(self, *args, starting_branch=None, creator=None, **kwargs):
        """Init the class."""
        super().__init__(*args, **kwargs)
        self.starting_branch = starting_branch
        self.creator = creator

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        """get_absolute_urls provide a url to access this branch's view."""
        return reverse("plugins:nautobot_version_control:branch", args=[self.name])

    @property
    def present_in_database(self):
        """present_in_database returns whether the branch exists in the database."""
        # determines `editing` flag in forms
        return Branch.objects.filter(name=self.name).exists()

    @property
    def active(self):
        """Returns true if the branch is the active branch."""
        return self.name == active_branch()

    @property
    def ahead_behind(self):
        """
        ahead_behind computes the ahead/behind. Ahead represents the number of commits since the ancestor of main and this
        branch. Behind represents how many commits are on main that have diverged passed this branch.
        :return: ahead/behind string.
        """
        merge_base = Commit.merge_base(self.name, DOLT_DEFAULT_BRANCH)
        merge_base_commit = Commit.objects.using(db_for_commit(self.hash)).get(commit_hash=merge_base)
        main_hash = Branch.objects.get(name=DOLT_DEFAULT_BRANCH).hash

        ahead = Commit.objects.filter(date__gt=merge_base_commit.date).using(db_for_commit(self.hash)).count()
        behind = Commit.objects.filter(date__gt=merge_base_commit.date).using(db_for_commit(main_hash)).count()

        return f"{ahead} ahead / {behind} behind"

    @property
    def created_by(self):
        """created_by returns the branch author."""
        m = self._branch_meta()
        return m.author if m else None

    @property
    def created_at(self):
        """created_at returns the datetime the branch was created."""
        m = self._branch_meta()
        return m.created if m else None

    @property
    def source_branch(self):
        """source_branch returns the name of the branch that this branch was originally checked out on."""
        m = self._branch_meta()
        return m.source_branch if m else None

    def checkout(self):
        """Checkout performs a checkout operation to this branch making it the active_branch."""
        with connection.cursor() as cursor:
            cursor.execute(f"""SELECT dolt_checkout("{self.name}") FROM dual;""")

    def _branch_meta(self):
        try:
            return BranchMeta.objects.get(branch=self.name)
        except ObjectDoesNotExist:
            return None

    def head(self):
        """Head returns the most recent commit for this branch as an object."""
        return Commit.objects.get(commit_hash=self.hash)

    def merge(self, merge_branch, user=None, squash=False):
        """
        Merge performs a merge operation between this branch and the merge_branch
        :param merge_branch: The branch to merge with
        :param user: The User object to associate the merge with
        :param squash: Whether or not to squash the merge thereby making it one commit
        :return:
        """
        author = author_from_user(user)
        self.checkout()
        with connection.cursor() as cursor:
            cursor.execute("SET dolt_force_transaction_commit = 1;")
            if squash:
                cursor.execute(
                    f"""SELECT dolt_merge(
                        '--squash',
                        '{merge_branch}'
                    ) FROM dual;"""
                )
            else:
                cursor.execute(
                    f"""SELECT dolt_merge(
                        '--no-ff',
                        '{merge_branch}'
                    ) FROM dual;"""
                )
            success = cursor.fetchone()[0] == 1
            if success:
                # only commit merged data on success
                msg = f"""merged "{merge_branch}" into "{self.name}"."""
                cursor.execute(
                    f"""SELECT dolt_commit(
                        '--all',
                        '--allow-empty',
                        '--message', '{msg}',
                        '--author', '{author}'
                    ) FROM dual;"""
                )
            else:
                cursor.execute("SELECT dolt_merge('--abort') FROM dual;")  # nosec
                raise DoltError(
                    format_html(
                        "{}",
                        mark_safe(
                            f"""Merging <strong>{merge_branch}</strong> into <strong>{self}</strong> created merge conflicts. Resolve merge conflicts to reattempt the merge."""
                        ),
                    )
                )

    def save(self, *args, **kwargs):
        """Save overrides the model save method."""
        with connection.cursor() as cursor:
            cursor.execute(
                f"""INSERT INTO dolt_branches (name,hash) VALUES ('{self.name}',hashof('{self.starting_branch}'));"""  # nosec
            )


@receiver(pre_delete, sender=Branch)
def delete_branch_pre_hook(sender, instance, using, **kwargs):  # pylint: disable=W0613
    """
    delete_branch_pre_hook intercepts the pre_delete signal for a branch. It returns an error if a branch that is about
    to be delete has pull requests that have not been deleted before.
    """
    # search the pull requests models for the same branch
    prs = PullRequest.objects.filter(Q(source_branch=instance.name) | Q(destination_branch=instance.name))

    if len(prs) > 0:
        pr_list = ",".join([f'"{pr}"' for pr in prs])
        raise DoltError(f"Must delete existing pull request(s): [{pr_list}] before deleting branch {instance.name}")


class BranchMeta(models.Model):
    """
    BranchMeta class has a 1:1 relation with a Branch. It represents internal of a branch that can't be represented in
    the dolt_branches system table.
    """

    branch = models.TextField(primary_key=True)
    source_branch = models.TextField()
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    created = models.DateTimeField(auto_now_add=True, blank=True, null=True)

    class Meta:
        """Meta class."""

        # table name cannot start with "dolt"
        db_table = "nautobot_version_control_branchmeta"


#
# Commits
#


class Commit(DoltSystemTable):
    """Commit represents a Dolt Commit primitive."""

    commit_hash = models.TextField(primary_key=True)
    committer = models.TextField()
    email = models.TextField()
    date = models.DateTimeField()
    message = models.TextField()

    class Meta:
        """Meta class."""

        managed = False
        db_table = "dolt_log"
        verbose_name_plural = "commits"

    def __str__(self):
        return self.commit_hash

    def get_absolute_url(self):
        """get_absolute_url returns a link to a view that displays the commit's info."""
        return reverse("plugins:nautobot_version_control:commit", args=[self.commit_hash])

    @staticmethod
    def merge_base(left, right):
        """merge_base returns the ancestor commit between two commits."""
        with connection.cursor() as c:
            # author credentials not set
            c.execute(f"SELECT DOLT_MERGE_BASE('{left}', '{right}') FROM dual;")
            return c.fetchone()[0]

    @staticmethod
    def revert(commits, user):
        """revert executes a revert command on a commit which undoes it from the commit log."""
        args = ", ".join([f"'{c}'" for c in commits])
        author = author_from_user(user)
        args += f", '--author', '{author}'"
        with connection.cursor() as c:
            c.execute(f"SELECT DOLT_REVERT({args}) FROM dual;")
            return c.fetchone()[0]

    @property
    def short_message(self):
        """short_message truncates a commit message."""
        split = self.message.split(";")
        return split[0] + f". Total number of changes: {len(split)}"

    @property
    def present_in_database(self):
        """present_in_database returns whether a commit object exists in the Dolt database."""
        # determines `editing` flag in forms
        return Commit.objects.filter(commit_hash=self.commit_hash).exists()

    @property
    def parent_commits(self):
        """parent_commits returns the hashes of the commit ancestor."""
        return CommitAncestor.objects.filter(commit_hash=self.commit_hash).values_list("parent_hash", flat=True)

    def save(self, *args, using="default", user=None, **kwargs):  # pylint: disable=W0221
        """save overrides the Django model save behavior and perform a commit on the database."""
        msg = self.message.replace('"', "")
        author = author_from_user(user)
        conn = connections[using]
        with conn.cursor() as cursor:
            cursor.execute(
                f"""
            SELECT dolt_commit(
                '--all',
                '--allow-empty',
                '--message', "{msg}",
                '--author', "{author}")
            FROM dual;"""
            )


class CommitAncestor(DoltSystemTable):
    """CommitAncestor models the set of ancestors or parents that precede a Commit."""

    commit_hash = models.TextField(primary_key=True)
    parent_hash = models.TextField()  # primary_key=True
    parent_index = models.IntegerField()  # primary_key=True

    class Meta:
        managed = False
        db_table = "dolt_commit_ancestors"
        verbose_name_plural = "commit_ancestors"

    def __str__(self):
        return f"{self.commit_hash} ancestor[{self.parent_index}]: {self.parent_hash}"

    def save(self, *args, **kwargs):
        """We override and prevent save as the dolt_commit_ancestors system table should not be modified."""
        # todo(andy): throw exception?
        return


#
# Conflicts
#


class Conflicts(DoltSystemTable):
    """Conflicts represents the dolt_conflicts system table and the conflicts from a merge it contains."""

    table = models.TextField(primary_key=True)
    num_conflicts = models.IntegerField()

    class Meta:
        managed = False
        db_table = "dolt_conflicts"
        verbose_name_plural = "conflicts"

    def __str__(self):
        return f"{self.table} ({self.num_conflicts})"


class ConstraintViolations(DoltSystemTable):
    """Foreign Key and Unique Key Constraint Violations."""

    table = models.TextField(primary_key=True)
    num_violations = models.IntegerField()

    class Meta:
        managed = False
        db_table = "dolt_constraint_violations"
        verbose_name_plural = "constraint violations"

    def __str__(self):
        return f"{self.table} ({self.num_violations})"


#
# Pull Requests
#


@extras_features(
    "webhooks",
)
class PullRequest(BaseModel):
    """PullRequest models a pull request between two branches."""

    OPEN = 0
    MERGED = 1
    CLOSED = 2
    PR_STATE_CHOICES = [
        (OPEN, "Open"),
        (MERGED, "Merged"),
        (CLOSED, "Closed"),
    ]

    title = models.CharField(max_length=240)
    state = models.IntegerField(choices=PR_STATE_CHOICES, default=OPEN)
    # can't create Foreign Key to dolt_branches table :(
    source_branch = models.TextField()
    destination_branch = models.TextField()
    description = models.TextField(blank=True, null=True)
    creator = models.ForeignKey(User, on_delete=CASCADE)
    created_at = models.DateTimeField(auto_now_add=True, blank=True, null=True)

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        # table name cannot start with "dolt"
        db_table = "nautobot_version_control_pull_request"
        verbose_name_plural = "pull requests"

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        """get_absolute_url returns a url to render a view of the pull request."""
        return reverse("plugins:nautobot_version_control:pull_request", args=[self.id])

    def get_src_dest_branches(self):
        """get_src_dest_branches returns a tuple of the src and destination branches."""
        src = Branch.objects.get(name=self.source_branch)
        dest = Branch.objects.get(name=self.destination_branch)
        return src, dest

    @property
    def open(self):
        """opens returns whether a pull request is open."""
        return self.state == PullRequest.OPEN

    @property
    def status(self):  # pylint: disable=R0911
        """
        The status of a PullRequest is determined by considering
        both the PullRequest and its PullRequestReviews.
        PRs in a closed or merged state have the corresponding status.
        An open PR's state is determined by the last non-comment review.
        """
        if self.state == PullRequest.CLOSED:
            return "closed"
        if self.state == PullRequest.MERGED:
            return "merged"

        pr_reviews = PullRequestReview.objects.filter(pull_request=self.pk)
        if not pr_reviews:
            return "open"

        # get the most recent review that approved or blocked
        decision = pr_reviews.order_by("-reviewed_at").first()
        if not decision or (decision and decision.state == PullRequestReview.COMMENTED):
            # all PRs are "comments"
            return "in-review"
        if decision.state == PullRequestReview.APPROVED:
            return "approved"
        if decision.state == PullRequestReview.BLOCKED:
            return "blocked"
        return "unknown"  # unreachable

    @property
    def commits(self):
        """commits returns a queryset of Commit objects that come after the ancestor between the src and des branch."""
        merge_base = Commit.objects.get(commit_hash=Commit.merge_base(self.source_branch, self.destination_branch))
        db = db_for_commit(Branch.objects.get(name=self.source_branch).hash)
        return Commit.objects.filter(date__gt=merge_base.date).using(db)

    @property
    def num_commits(self):
        """num_commits returns the number of commits that are considered in the pull requests."""
        return self.commits.count()

    @property
    def num_reviews(self):
        """num_reviews returns the number of PullRRequestReview(s) created on top of the PR."""
        return PullRequestReview.objects.filter(pull_request=self.pk).count()

    @property
    def summary_description(self):
        """summary_description returns a small summary of the pull request action."""
        return f"""Merging {self.num_commits} commits from "{self.source_branch}" into "{self.destination_branch}" """

    def merge(self, user=None, squash=False):
        """merge executes a merge between a destination and src branch."""
        try:
            src = Branch.objects.get(name=self.source_branch)
            dest = Branch.objects.get(name=self.destination_branch)
            dest.merge(src, user=user, squash=squash)
        except ObjectDoesNotExist as e:
            raise DoltError(f"error merging Pull Request {self}: {e}") from e
        self.state = PullRequest.MERGED
        self.save()


@extras_features(
    "webhooks",
)
class PullRequestReview(BaseModel):
    """PullRequestReview represents a comments, approval, or block on a pull request."""

    COMMENTED = 0
    APPROVED = 1
    BLOCKED = 2
    REVIEW_STATE_CHOICES = [
        (COMMENTED, "Commented"),
        (APPROVED, "Approved"),
        (BLOCKED, "Blocked"),
    ]

    pull_request = models.ForeignKey(PullRequest, on_delete=CASCADE)
    reviewer = models.ForeignKey(User, on_delete=CASCADE)
    reviewed_at = models.DateTimeField(auto_now_add=True, blank=True, null=True)
    state = models.IntegerField(choices=REVIEW_STATE_CHOICES, null=True)
    summary = models.TextField(blank=True, null=True)

    class Meta:
        # table name cannot start with "dolt"
        db_table = "nautobot_version_control_pull_request_review"
        verbose_name_plural = "pull request reviews"

    def __str__(self):
        return f""""{self.pull_request}" reviewed by {self.reviewer}"""

    def get_absolute_url(self):
        """get_absolute_url returns a link to a view of a pull request review."""
        return reverse("plugins:nautobot_version_control:pull_request", args=[self.pull_request.id])



# Generated by Django 3.1.13 on 2021-08-12 20:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("nautobot_version_control", "0004_constraintviolations"),
    ]

    operations = [
        migrations.CreateModel(
            name="PullRequest",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("title", models.CharField(max_length=240)),
                ("state", models.IntegerField(default=0)),
                ("source_branch", models.TextField()),
                ("destination_branch", models.TextField()),
                ("description", models.TextField()),
                ("created_at", models.DateField(auto_now_add=True, null=True)),
                (
                    "creator",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "pull requests",
                "db_table": "nautobot_version_control_pull_request",
            },
        ),
        migrations.CreateModel(
            name="PullRequestReview",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("reviewed_at", models.DateTimeField(auto_now_add=True, null=True)),
                ("state", models.IntegerField(null=True)),
                ("summary", models.TextField()),
                (
                    "pull_request",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="nautobot_version_control.pullrequest",
                    ),
                ),
                (
                    "reviewer",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "pull request reviews",
                "db_table": "nautobot_version_control_pull_request_review",
            },
        ),
    ]



"""Dunctions.py represents custom SQL functions that do not fit the django orm."""

from django.db.models.expressions import Func, Value
from django.db.models.fields.json import JSONField


# pylint: disable=W0223
class JSONObject(Func):
    """JSONObject represents the json_object mysql function."""

    function = "JSON_OBJECT"
    output_field = JSONField()

    def __init__(self, **fields):
        """Add expressions."""
        expressions = []
        for key, value in fields.items():
            expressions.extend((Value(key), value))
        super().__init__(*expressions)



"""Diffs.py contains a set of utilities for producing Dolt diffs."""

from django.contrib.contenttypes.models import ContentType
from django.db import connection
from django.db import models
from django.db.models.expressions import RawSQL

from nautobot.circuits import tables as circuits_tables
from nautobot.dcim.tables import cables, devices, devicetypes, power, racks, sites
from nautobot.extras import tables as extras_tables
from nautobot.ipam import tables as ipam_tables
from nautobot.tenancy import tables as tenancy_tables
from nautobot.virtualization import tables as virtualization_tables

from nautobot_version_control.dynamic.diff_factory import DiffListViewFactory
from nautobot_version_control.models import Commit
from nautobot_version_control.utils import db_for_commit

from . import diff_table_for_model, register_diff_tables


def three_dot_diffs(from_commit=None, to_commit=None):
    """three_dot_diffs returns a diff between the ancestor of from_to_commit with to_commit."""
    if not (from_commit and to_commit):
        raise ValueError("must specify both a to_commit and from_commit")
    merge_base = Commit.merge_base(from_commit, to_commit)
    return two_dot_diffs(from_commit=merge_base, to_commit=to_commit)


def two_dot_diffs(from_commit=None, to_commit=None):
    """two_dot_diffs returns the diff between from_commit and to_commit via the dolt diff table interface."""
    if not (from_commit and to_commit):
        raise ValueError("must specify both a to_commit and from_commit")

    diff_results = []
    for content_type in ContentType.objects.all():
        model = content_type.model_class()
        if not model:
            continue
        if not diff_table_for_model(model):
            continue

        ct_meta = content_type.model_class()._meta
        tbl_name = ct_meta.db_table
        verbose_name = str(ct_meta.verbose_name.capitalize())

        to_queryset = (
            content_type.model_class()
            .objects.filter(
                pk__in=RawSQL(  # nosec
                    f"""SELECT to_id FROM dolt_commit_diff_{tbl_name}
                        WHERE to_commit = %s AND from_commit = %s""",
                    (to_commit, from_commit),
                )
            )
            .annotate(
                # Annotate each row with a JSON-ified diff
                diff=RawSQL(  # nosec
                    f"""SELECT JSON_OBJECT("root", "to", {json_diff_fields(tbl_name)})
                        FROM dolt_commit_diff_{tbl_name}
                        WHERE to_commit = %s AND from_commit = %s
                        AND to_id = {tbl_name}.id """,
                    (to_commit, from_commit),
                    output_field=models.JSONField(),
                )
            )
            # "time-travel" query the database at `to_commit`
            .using(db_for_commit(to_commit))
        )

        from_queryset = (
            content_type.model_class()
            .objects.filter(
                # add the `diff_type = 'removed'` clause, because we only want deleted
                # rows in this queryset. modified rows come from the `to_queryset`
                pk__in=RawSQL(  # nosec
                    f"""SELECT from_id FROM dolt_commit_diff_{tbl_name}
                        WHERE to_commit = %s AND from_commit = %s AND diff_type = 'removed' """,
                    (to_commit, from_commit),
                )
            )
            .annotate(
                # Annotate each row with a JSON-ified diff
                diff=RawSQL(  # nosec
                    f"""SELECT JSON_OBJECT("root", "from", {json_diff_fields(tbl_name)})
                        FROM dolt_commit_diff_{tbl_name}
                        WHERE to_commit = %s AND from_commit = %s
                        AND from_id = {tbl_name}.id """,
                    (to_commit, from_commit),
                    output_field=models.JSONField(),
                )
            )
            # "time-travel" query the database at `from_commit`
            .using(db_for_commit(from_commit))
        )
        diff_rows = sorted(list(to_queryset) + list(from_queryset), key=lambda d: d.pk)
        if len(diff_rows) == 0:
            continue

        diff_view_table = DiffListViewFactory(content_type).get_table_model()
        diff_results.append(
            {
                "name": f"{verbose_name} Diffs",
                "table": diff_view_table(diff_rows),
                **diff_summary_for_table(tbl_name, from_commit, to_commit),
            }
        )
    return diff_results


def diff_summary_for_table(table, from_commit, to_commit):
    """diff_summary_for_table returns the diff summary for table, for the commits from_commit and to_commit."""
    summary = {
        "added": 0,
        "modified": 0,
        "removed": 0,
    }
    with connection.cursor() as cursor:
        cursor.execute(
            f"""SELECT diff_type, count(diff_type) FROM dolt_commit_diff_{table}  # nosec
                WHERE to_commit = %s AND from_commit = %s
                GROUP BY diff_type ORDER BY diff_type""",  # nosec
            (to_commit, from_commit),
        )
        for diff_type, count in cursor.fetchall():
            summary[diff_type] = count
    return summary


def json_diff_fields(tbl_name):
    """
    json_diff_fields returns all of the column names for a model
    and turns them into to_ and from_ fields.
    """
    with connection.cursor() as cursor:
        cursor.execute(f"DESCRIBE dolt_commit_diff_{tbl_name}")
        cols = cursor.fetchall()
    pairs = (f"'{c[0]}', dolt_commit_diff_{tbl_name}.{c[0]}" for c in cols)
    return ", ".join(pairs)


register_diff_tables(
    {
        "circuits": {
            "circuit": circuits_tables.CircuitTable,
            # "circuittermination": None,
            "circuittype": circuits_tables.CircuitTypeTable,
            "provider": circuits_tables.ProviderTable,
        },
        "dcim": {
            # "baseinterface": devices.BaseInterfaceTable,
            "cable": cables.CableTable,
            # "cabletermination": devices.CableTerminationTable,
            # "componenttemplate": devicetypes.ComponentTemplateTable,
            "consoleport": devices.ConsolePortTable,
            "consoleporttemplate": devicetypes.ConsolePortTemplateTable,
            "consoleserverport": devices.ConsoleServerPortTable,
            "consoleserverporttemplate": devicetypes.ConsoleServerPortTemplateTable,
            "device": devices.DeviceTable,
            "devicebay": devices.DeviceBayTable,
            "devicebaytemplate": devicetypes.DeviceBayTemplateTable,
            # "devicecomponent": devices.DeviceComponentTable,
            # "deviceconsoleport": devices.DeviceConsolePortTable,
            # "deviceconsoleserverport": devices.DeviceConsoleServerPortTable,
            # "devicedevicebay": devices.DeviceDeviceBayTable,
            # "devicefrontport": devices.DeviceFrontPortTable,
            # "deviceimport": devices.DeviceImportTable,
            # "deviceinterface": devices.DeviceInterfaceTable,
            # "deviceinventoryitem": devices.DeviceInventoryItemTable,
            # "devicepoweroutlet": devices.DevicePowerOutletTable,
            # "devicepowerport": devices.DevicePowerPortTable,
            # "devicerearport": devices.DeviceRearPortTable,
            "devicerole": devices.DeviceRoleTable,
            "devicetype": devicetypes.DeviceTypeTable,
            "frontport": devices.FrontPortTable,
            "frontporttemplate": devicetypes.FrontPortTemplateTable,
            "interface": devices.InterfaceTable,
            "interfacetemplate": devicetypes.InterfaceTemplateTable,
            "inventoryitem": devices.InventoryItemTable,
            "manufacturer": devicetypes.ManufacturerTable,
            # "pathendpoint": devices.PathEndpointTable,
            "platform": devices.PlatformTable,
            "powerfeed": power.PowerFeedTable,
            "poweroutlet": devices.PowerOutletTable,
            "poweroutlettemplate": devicetypes.PowerOutletTemplateTable,
            "powerpanel": power.PowerPanelTable,
            "powerport": devices.PowerPortTable,
            "powerporttemplate": devicetypes.PowerPortTemplateTable,
            "rack": racks.RackTable,
            # "rackdetail": racks.RackDetailTable,
            "rackgroup": racks.RackGroupTable,
            "rackreservation": racks.RackReservationTable,
            "rackrole": racks.RackRoleTable,
            "rearport": devices.RearPortTable,
            "rearporttemplate": devicetypes.RearPortTemplateTable,
            "region": sites.RegionTable,
            "site": sites.SiteTable,
            "virtualchassis": devices.VirtualChassisTable,
        },
        "extras": {
            "secret": extras_tables.SecretTable,
            "secretsgroup": extras_tables.SecretsGroupTable,
        },
        "ipam": {
            "aggregate": ipam_tables.AggregateTable,
            "ipaddress": ipam_tables.IPAddressTable,
            "prefix": ipam_tables.PrefixTable,
            "rir": ipam_tables.RIRTable,
            "role": ipam_tables.RoleTable,
            "routetarget": ipam_tables.RouteTargetTable,
            "service": ipam_tables.ServiceTable,
            "vlan": ipam_tables.VLANTable,
            "vlangroup": ipam_tables.VLANGroupTable,
            "vrf": ipam_tables.VRFTable,
        },
        "tenancy": {
            "tenantgroup": tenancy_tables.TenantGroupTable,
            "tenant": tenancy_tables.TenantTable,
        },
        "virtualization": {
            "cluster": virtualization_tables.ClusterTypeTable,
            "clustergroup": virtualization_tables.ClusterGroupTable,
            "clustertype": virtualization_tables.ClusterTable,
            "vminterface": virtualization_tables.VMInterfaceTable,
        },
    }
)



# Generated by Django 3.1.13 on 2021-08-18 17:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nautobot_version_control", "0006_auto_20210817_0251"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pullrequest",
            name="created_at",
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
    ]



"""middleware.py contains the middleware add-ons needed for the Version Control plugin to work."""

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import m2m_changed, post_save, pre_delete
from django.http import HttpResponse
from django.shortcuts import redirect
from django.utils.safestring import mark_safe

from nautobot.extras.models.change_logging import ObjectChange

from nautobot_version_control.constants import (
    DOLT_BRANCH_KEYWORD,
    DOLT_DEFAULT_BRANCH,
)
from nautobot_version_control.models import Branch, Commit
from nautobot_version_control.utils import DoltError


def dolt_health_check_intercept_middleware(get_response):
    """
    Intercept health check calls and disregard.
    TODO: fix health-check and remove
    """

    def middleware(request):
        if "/health" in request.path:
            return HttpResponse(status=201)
        return get_response(request)

    return middleware


class DoltBranchMiddleware:
    """DoltBranchMiddleware keeps track of which branch the dolt database is on."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """Override __call__."""
        return self.get_response(request)

    def process_view(self, request, view_func, view_args, view_kwargs):  # pylint: disable=R0201
        """
        process_view maintains the dolt branch session cookie and verifies authentication. It then returns the
        view that needs to be rendered.
        """
        # Check whether the desired branch was passed in as a querystring
        query_string_branch = request.GET.get(DOLT_BRANCH_KEYWORD, None)
        if query_string_branch is not None:
            # update the session Cookie
            request.session[DOLT_BRANCH_KEYWORD] = query_string_branch
            return redirect(request.path)

        branch = DoltBranchMiddleware.get_branch(request)
        try:
            branch.checkout()
        except Exception as e:
            msg = f"could not checkout branch {branch}: {str(e)}"
            messages.error(request, mark_safe(msg))

        try:
            return view_func(request, *view_args, **view_kwargs)
        except DoltError as e:
            messages.error(request, mark_safe(e))
            return redirect(request.path)

    @staticmethod
    def get_branch(request):
        """get_branch returns the Branch object of the branch stored in the session cookie."""
        # lookup the active branch in the session cookie
        requested = branch_from_request(request)
        try:
            return Branch.objects.get(pk=requested)
        except ObjectDoesNotExist:
            messages.warning(
                request,
                mark_safe(f"""<div class="text-center">branch not found: {requested}</div>"""),  # nosec
            )
            request.session[DOLT_BRANCH_KEYWORD] = DOLT_DEFAULT_BRANCH
            return Branch.objects.get(pk=DOLT_DEFAULT_BRANCH)


class DoltAutoCommitMiddleware:
    """
    DoltAutoCommitMiddleware calls the AutoDoltCommit class on a request.
    - adapted from nautobot.extras.middleware.ObjectChangeMiddleware.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """Override call."""
        # Process the request with auto-dolt-commit enabled
        with AutoDoltCommit(request):
            return self.get_response(request)


class AutoDoltCommit:
    """
    AutoDoltCommit handles automatic dolt commits on the case than objects is created or deleted.
    - adapted from `nautobot.extras.context_managers`.
    """

    def __init__(self, request):
        self.request = request
        self.commit = False
        self.changes_for_db = {}

    def __enter__(self):
        # Connect our receivers to the post_save and post_delete signals.
        post_save.connect(self._handle_update, dispatch_uid="dolt_commit_update")
        m2m_changed.connect(self._handle_update, dispatch_uid="dolt_commit_update")
        pre_delete.connect(self._handle_delete, dispatch_uid="dolt_commit_delete")

    def __exit__(self, type, value, traceback):  # pylint: disable=W0622
        if self.commit:
            self.make_commits()

        # Disconnect change logging signals. This is necessary to avoid recording any errant
        # changes during test cleanup.
        post_save.disconnect(self._handle_update, dispatch_uid="dolt_commit_update")
        m2m_changed.disconnect(self._handle_update, dispatch_uid="dolt_commit_update")
        pre_delete.disconnect(self._handle_delete, dispatch_uid="dolt_commit_delete")

    def _handle_update(self, sender, instance, **kwargs):  # pylint: disable=W0613
        """Fires when an object is created or updated."""
        if isinstance(instance, ObjectChange):
            # ignore ObjectChange instances
            return

        msg = self.change_msg_for_update(instance, kwargs)
        self.collect_change(instance, msg)
        self.commit = True

    def _handle_delete(self, sender, instance, **kwargs):  # pylint: disable=W0613
        """Fires when an object is deleted."""
        if isinstance(instance, ObjectChange):
            # ignore ObjectChange instances
            return

        msg = self.change_msg_for_delete(instance)
        self.collect_change(instance, msg)
        self.commit = True

    def make_commits(self):
        """make_commits creates and saves a Commit object."""
        for db, msgs in self.changes_for_db.items():
            msg = "; ".join(msgs)
            Commit(message=msg).save(
                user=self.request.user,
                using=db,
            )

    def collect_change(self, instance, msg):
        """collect_change stores changes messages for each db."""
        db = self.database_from_instance(instance)
        if db not in self.changes_for_db:
            self.changes_for_db[db] = []
        self.changes_for_db[db].append(msg)

    @staticmethod
    def database_from_instance(instance):
        """database_from_instance returns a database from an instance type."""
        return instance._state.db  # pylint: disable=W0212

    @staticmethod
    def change_msg_for_update(instance, kwargs):
        """Generates a commit message for create or update."""
        created = "created" in kwargs and kwargs["created"]
        verb = "Created" if created else "Updated"
        return f"""{verb} {instance._meta.verbose_name} "{instance}" """

    @staticmethod
    def change_msg_for_delete(instance):
        """Generates a commit message for delete."""
        return f"""Deleted {instance._meta.verbose_name} "{instance}" """


def branch_from_request(request):
    """
    Returns the active branch from a request
    :param request: A django request
    :return: Branch name
    """
    if DOLT_BRANCH_KEYWORD in request.session:
        return request.session.get(DOLT_BRANCH_KEYWORD)
    if DOLT_BRANCH_KEYWORD in request.headers:
        return request.headers.get(DOLT_BRANCH_KEYWORD)
    return DOLT_DEFAULT_BRANCH



"""Merge.py contains utilities for the merging two branches and detecting any conflicts."""

import json

from django.core.exceptions import ObjectDoesNotExist
from django.db import connection
from django.db.models import Sum
from django.contrib.contenttypes.models import ContentType
from django.utils.safestring import mark_safe

from nautobot_version_control.models import (
    Branch,
    Conflicts,
    ConstraintViolations,
    Commit,
)
from nautobot_version_control.utils import author_from_user, query_on_branch
from nautobot_version_control.tables import (
    ConflictsTable,
    ConstraintViolationsTable,
)


# TODO: this file should be named "conflicts.py"


def get_conflicts_count_for_merge(src, dest):
    """
    Gather a merge-candidate for `src` and `dest`, then return Conflicts created by the merge.

    TODO: currently we return conflicts summary,
        we need granular row-level conflicts and
        constraint violations.
    """
    try:
        mc = get_or_make_merge_candidate(src, dest)
        with query_on_branch(mc):
            c = Conflicts.objects.all().aggregate(Sum("num_conflicts"))
            v = ConstraintViolations.objects.all().aggregate(Sum("num_violations"))
            num_conflicts = c["num_conflicts__sum"] if c["num_conflicts__sum"] else 0
            num_violations = v["num_violations__sum"] if v["num_violations__sum"] else 0
            return num_conflicts + num_violations
    except Exception:
        # best effort
        # todo: fix dolt merge bug
        return 0


def get_conflicts_for_merge(src, dest):
    """
    Gather a merge-candidate for `src` and `dest`, then return Conflicts created by the merge.

    TODO: currently we return conflicts summary,
        we need granular row-level conflicts and
        constraint violations.
    """
    try:
        mc = get_or_make_merge_candidate(src, dest)
        with query_on_branch(mc):
            conflicts = MergeConflicts(src, dest)
            return {
                "summary": conflicts.make_conflict_summary_table(),
                "conflicts": conflicts.make_conflict_table(),
                "violations": conflicts.make_constraint_violations_table(),
            }

    except Exception:
        # best effort
        # todo: fix dolt merge bug
        return {}


def merge_candidate_exists(src, dest):
    """merge_candidate_exists returns true if there exist a merge_candidate branch between src and dest."""
    name = _merge_candidate_name(src, dest)
    try:
        mc = Branch.objects.get(name=name)
        return merge_candidate_is_fresh(mc, src, dest)
    except Branch.DoesNotExist:
        return False


def merge_candidate_is_fresh(mc, src, dest):
    """A merge candidate (MC) is considered "fresh" if the source and destination branches used to create the MC are unchanged since the MC was created."""
    if not mc:
        return False
    src_stable = Commit.merge_base(mc, src) == src.hash
    dest_stable = Commit.merge_base(mc, dest) == dest.hash
    return src_stable and dest_stable


def get_merge_candidate(src, dest):
    """get_merge_candidate returns a merge candidate branch if it exists."""
    if merge_candidate_exists(src, dest):
        name = _merge_candidate_name(src, dest)
        return Branch.objects.get(name=name)
    return None


def make_merge_candidate(src, dest):
    """make_merge_candidate create a merge candidate branch between src and dest."""
    name = _merge_candidate_name(src, dest)
    # force updates the merge-candidate branch
    Branch(name=name, starting_branch=dest).save()
    with connection.cursor() as cursor:
        cursor.execute("SET @@dolt_force_transaction_commit = 1;")
        cursor.execute("""SELECT dolt_checkout(%s) FROM dual;""", [name])
        cursor.execute("""SELECT dolt_merge(%s) FROM dual;""", [src])
        cursor.execute("""SELECT dolt_add("-A") FROM dual;""")
        msg = f"""creating merge candidate with src: "{src}" and dest: "{dest}"."""
        cursor.execute(
            f"""SELECT dolt_commit(
                    '--force',
                    '--all',
                    '--allow-empty',
                    '--message', '{msg}',
                    '--author', '{author_from_user(None)}') FROM dual;"""
        )
    return Branch.objects.get(name=name)


def get_or_make_merge_candidate(src, dest):
    """Gets or creates a merge candidate branch between src and dest."""
    mc = get_merge_candidate(src, dest)
    if not mc:
        mc = make_merge_candidate(src, dest)
    return mc


def _merge_candidate_name(src, dest):
    """Returns the formatted name of a merge candidate branch."""
    return f"xxx-merge-candidate--{src}--{dest}"


class MergeConflicts:
    """Must run under the mc branch."""

    def __init__(self, src, dest):
        """Inits the class vars."""
        self.src = src
        self.dest = dest
        self.model_map = self._model_map()

    @staticmethod
    def _model_map():
        return {ct.model_class()._meta.db_table: ct.model_class() for ct in ContentType.objects.all()}

    def make_conflict_summary_table(self):
        """Creates the conflict summary table for merge conflicts."""
        conflicts = Conflicts.objects.all()
        violations = ConstraintViolations.objects.all()
        summary = {
            c.table: {
                "model": self._model_from_table(c.table),
                "num_conflicts": c.num_conflicts,
            }
            for c in conflicts
        }
        for v in violations:
            if v.table not in summary:
                summary[v.table] = {"model": self._model_from_table(v.table)}
            summary[v.table]["num_violations"] = v.num_violations
        return list(summary.values())

    def make_conflict_table(self):
        """Create a table that represents conflicts on a table between src and dest."""
        rows = []
        for c in Conflicts.objects.all():
            rows.extend(self.get_rows_level_conflicts(c))
        return ConflictsTable(rows)

    def make_constraint_violations_table(self):
        """make_constraint_violations_table creates a table to store constraint violations between two tables."""
        rows = []
        for v in ConstraintViolations.objects.all():
            rows.extend(self.get_rows_level_violations(v))
        return ConstraintViolationsTable(rows)

    def get_rows_level_conflicts(self, conflict):
        """get_row_level_conflicts returns each conflict row in a table as a JSON object."""
        with connection.cursor() as cursor:
            # introspect table schema to query conflict data as json
            cursor.execute(f"DESCRIBE dolt_conflicts_{conflict.table}")
            fields = ",".join([f"'{tup[0]}', {tup[0]}" for tup in cursor.fetchall()])

            cursor.execute(
                f"""SELECT base_id, JSON_OBJECT({fields})
                    FROM dolt_conflicts_{conflict.table};"""
            )
            model_name = self._model_from_table(conflict.table)
            return [
                {
                    "model": model_name,
                    "id": self._object_name_from_id(conflict.table, tup[0]),
                    "conflicts": self._transform_conflicts_obj(tup[1]),
                }
                for tup in cursor.fetchall()
            ]

    def _transform_conflicts_obj(self, obj):
        if isinstance(obj, str):
            obj = json.loads(obj)
        obj2 = {}
        for k, _ in obj.items():
            prefix = "our_"
            if not k.startswith(prefix):
                continue
            suffix = k[len(prefix) :]  # noqa: E203
            ours = obj[f"our_{suffix}"]
            theirs = obj[f"their_{suffix}"]
            base = obj[f"base_{suffix}"]

            conflicted = ours not in (theirs, base)
            if conflicted:
                obj2[suffix] = {
                    f"{self.dest}": ours,
                    f"{self.src}": theirs,
                    "base": base,
                }
        return obj2

    def get_rows_level_violations(self, violation):
        """get_row_level_violations returns each constrain violation in a JSON row."""
        with connection.cursor() as cursor:
            rows = []
            model_name = self._model_from_table(violation.table)
            cursor.execute(
                mark_safe(
                    """SELECT id, violation_type, violation_info
                    FROM dolt_constraint_violations_{violation.table};"""
                )
            )
            for v_row in cursor.fetchall():
                obj_name = self._object_name_from_id(violation.table, v_row[0])
                rows.append(
                    {
                        "model": model_name,
                        "id": obj_name,
                        "violation_type": v_row[1],
                        "violations": self._fmt_violation(v_row, model_name, obj_name),
                    }
                )
            return rows

    def _model_from_table(self, tbl_name):
        model = self.model_map[tbl_name]
        return model._meta.verbose_name

    def _object_name_from_id(self, tbl_name, id_):
        try:
            model = self.model_map[tbl_name]
            obj = model.objects.get(id=id_)
            return str(obj)
        except ObjectDoesNotExist:
            return id_

    def _fmt_violation(self, v_row, model_name, obj_name):
        v_type = v_row[1]
        v_info = json.loads(v_row[2])

        if v_type == "foreign key":
            if "ReferencedTable" in v_info:
                rt = v_info["ReferencedTable"]
                ref_model_name = self._model_from_table(rt)
                return f"""
                    The {model_name} "{obj_name}" references a
                    missing "{ref_model_name}" object
                """

        elif v_type == "unique index":
            if "Columns" in v_info:
                return f"""
                    The {model_name} "{obj_name}" violates a
                    uniqueness constraint defined over the
                    columns {v_info["Columns"]}
                """
        return "Unknown constraint violation"



"""cleanup_data.py."""

from django.core.management.base import BaseCommand
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from nautobot.extras.models import Status


class Command(BaseCommand):
    """Cleanup Database after migrations."""

    help = "Cleanup Database after migrations."

    def handle(self, *args, **kwargs):
        """override handle."""
        Status.objects.all().delete()
        ContentType.objects.all().delete()
        Permission.objects.all().delete()



"""views.py implements views for the api."""

from rest_framework.routers import APIRootView
from nautobot.extras.api.views import CustomFieldModelViewSet

from nautobot_version_control import filters
from nautobot_version_control.models import Branch, Commit, PullRequest, PullRequestReview

from . import serializers


class VCSRootView(APIRootView):
    """VCS API root view."""

    def get_view_name(self):
        """returns the name of the view."""
        return "VCS"


#
# Branches
#


class BranchViewSet(CustomFieldModelViewSet):
    """BranchViewSet render a view for the Branch model."""

    queryset = Branch.objects.all()
    serializer_class = serializers.BranchSerializer
    filterset_class = filters.BranchFilterSet


#
# Commits
#


class CommitViewSet(CustomFieldModelViewSet):
    """CommitViewSet render a view for the Commit model."""

    queryset = Commit.objects.all()
    serializer_class = serializers.CommitSerializer
    filterset_class = filters.CommitFilterSet


#
# Pull Requests
#


class PullRequestViewSet(CustomFieldModelViewSet):
    """PullRequestViewSet render a view for the PullRequest model."""

    queryset = PullRequest.objects.all()
    serializer_class = serializers.PullRequestSerializer
    filterset_class = filters.PullRequestFilterSet


#
# Pull Request Reviews
#


class PullRequestReviewViewSet(CustomFieldModelViewSet):
    """PullRequestReviewViewSet render a view for the PullRequestReviewV model."""

    queryset = PullRequestReview.objects.all()
    serializer_class = serializers.PullRequestReviewSerializer
    filterset_class = filters.PullRequestReviewFilterSet



"""Filters.py defines a set of Filters needed for each model defined in models.py."""

import django_filters
from django.db.models import Q

from nautobot.utilities.filters import BaseFilterSet
from nautobot_version_control.models import Branch, Commit, PullRequest, PullRequestReview


class BranchFilterSet(BaseFilterSet):
    """BranchFilterSet returns a filter for the Branch model."""

    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    class Meta:
        model = Branch
        fields = (
            "name",
            "hash",
            "latest_committer",
            "latest_committer_email",
            "latest_commit_date",
            "latest_commit_message",
        )

    def search(self, queryset, name, value):  # pylint: disable=unused-argument,no-self-use
        """
        search performs an ORM filter on the Branch model
        :param queryset: The Branch queryset
        :param name: The modelname
        :param value: The value to be searched for
        :return: A filtered queryset
        """
        value = value.strip()
        if not value:
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
            | Q(hash__icontains=value)
            | Q(latest_committer__icontains=value)
            | Q(latest_committer_email__icontains=value)
            | Q(latest_commit_date__icontains=value)
            | Q(latest_commit_message__icontains=value)
        )


class CommitFilterSet(BaseFilterSet):
    """CommitFilterSet returns a filter for the CommitFilterSet."""

    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    class Meta:
        model = Commit
        fields = (
            "commit_hash",
            "committer",
            "email",
            "date",
            "message",
        )

    def search(self, queryset, name, value):  # pylint: disable=unused-argument,no-self-use
        """
        search performs an ORM filter on the Commit model
        :param queryset: The Commit queryset
        :param name: The modelname
        :param value: The value to be searched for
        :return: A filtered queryset
        """
        value = value.strip()
        if not value:
            return queryset
        return queryset.filter(
            Q(commit_hash__icontains=value)
            | Q(committer__icontains=value)
            | Q(email__icontains=value)
            | Q(date__icontains=value)
            | Q(message__icontains=value)
        )


class PullRequestFilterSet(BaseFilterSet):
    """PullRequestFilterSet returns a filter for the PullRequest model."""

    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )

    class Meta:
        model = PullRequest
        fields = (
            "title",
            "state",
            "source_branch",
            "destination_branch",
            "description",
            "creator",
        )

    def search(self, queryset, name, value):  # pylint: disable=unused-argument,no-self-use
        """
        search performs an ORM filter on the PullRequestFilterSet model
        :param queryset: The PullRequestFilterSet queryset
        :param name: The modelname
        :param value: The value to be searched for
        :return: A filtered queryset
        """
        value = value.strip()
        if not value:
            return queryset
        return queryset.filter(
            Q(title__icontains=value)
            | Q(state__icontains=value)
            | Q(source_branch__icontains=value)
            | Q(destination_branch__icontains=value)
            | Q(description__icontains=value)
            | Q(creator__icontains=value)
        )


class PullRequestDefaultOpenFilterSet(PullRequestFilterSet):
    """PullRequestDefaultOpenFilterSet returns a filter for the PullRequest model where the default search is state=OPEN."""

    state = django_filters.MultipleChoiceFilter(choices=PullRequest.PR_STATE_CHOICES)

    def __init__(self, data, *args, **kwargs):
        if not data.get("state"):
            data = data.copy()
            data["state"] = PullRequest.OPEN
        super().__init__(data, *args, **kwargs)


class PullRequestReviewFilterSet(BaseFilterSet):
    """PullRequestReviewFilterSet returns a filter for the PullRequestReview model."""

    q = django_filters.CharFilter(
        method="search",
        label="Search",
    )
    state = django_filters.MultipleChoiceFilter(choices=PullRequest.PR_STATE_CHOICES)

    class Meta:
        model = PullRequestReview
        fields = (
            "pull_request",
            "reviewer",
            "state",
            "reviewed_at",
            "summary",
        )

    def search(self, queryset, name, value):  # pylint: disable=unused-argument,no-self-use
        """
        search performs an ORM filter on the PullRequestReviewFilterSet model
        :param queryset: The PullRequestReviewFilterSet queryset
        :param name: The modelname
        :param value: The value to be searched for
        :return: A filtered queryset
        """
        value = value.strip()
        if not value:
            return queryset
        return queryset.filter(
            Q(reviewer__icontains=value)
            | Q(reviewed_at__icontains=value)
            | Q(state__icontains=value)
            | Q(summary__icontains=value)
        )



"""urls.py contains the main routing for the Nautobot Version Control plugin."""


from django.urls import path

from nautobot_version_control import views

urlpatterns = [
    # Branches
    path("branches/", views.BranchListView.as_view(), name="branch_list"),
    path("branches/add/", views.BranchEditView.as_view(), name="branch_add"),
    path("branches/edit/", views.BranchBulkEditView.as_view(), name="branch_bulk_edit"),
    path(
        "branches/delete/",
        views.BranchBulkDeleteView.as_view(),
        name="branch_bulk_delete",
    ),
    path(
        "branches/<str:src>/merge/",
        views.BranchMergeFormView.as_view(),
        name="branch_merge",
    ),
    path(
        "branches/<str:src>/merge/<str:dest>",
        views.BranchMergePreView.as_view(),
        name="branch_merge_preview",
    ),
    path("branches/<str:pk>/", views.BranchView.as_view(), name="branch"),
    path(
        "branches/<str:pk>/checkout/",
        views.BranchCheckoutView.as_view(),
        name="branch_checkout",
    ),
    path("branches/<str:pk>/edit/", views.BranchEditView.as_view(), name="branch_edit"),
    # Commits
    path("commits/", views.CommitListView.as_view(), name="commit_list"),
    path("commits/add/", views.CommitEditView.as_view(), name="commit_add"),
    path(
        "commits/revert/",
        views.CommitRevertView.as_view(),
        name="commit_revert",
    ),
    path("commits/<str:pk>/", views.CommitView.as_view(), name="commit"),
    path("commits/<str:pk>/edit/", views.CommitEditView.as_view(), name="commit_edit"),
    path(
        "commits/<str:pk>/delete/",
        views.CommitDeleteView.as_view(),
        name="commit_delete",
    ),
    path(
        "dynamic/<str:from_commit>/<str:to_commit>/<str:app_label>/<str:model>/<str:pk>/",
        views.DiffDetailView.as_view(),
        name="diff_detail",
    ),
    # Diffs
    path("diffs/", views.ActiveBranchDiffs.as_view(), name="active_branch_diffs"),
    # Pull Requests
    path("pull-request/", views.PullRequestListView.as_view(), name="pull_request_list"),
    path(
        "pull-request/add/",
        views.PullRequestEditView.as_view(),
        name="pull_request_add",
    ),
    path(
        "pull-request/delete/",
        views.PullRequestBulkDeleteView.as_view(),
        name="pullrequest_bulk_delete",
    ),
    path(
        "pull-request/<str:pk>/",
        views.PullRequestDiffView.as_view(),
        name="pull_request",
    ),
    path(
        "pull-request/<str:pk>/edit",
        views.PullRequestEditView.as_view(),
        name="pull_request_edit",
    ),
    path(
        "pull-request/<str:pk>/merge",
        views.PullRequestMergeView.as_view(),
        name="pull_request_merge",
    ),
    path(
        "pull-request/<str:pk>/close",
        views.PullRequestCloseView.as_view(),
        name="pull_request_close",
    ),
    path(
        "pull-request/<str:pk>/conflicts",
        views.PullRequestConflictView.as_view(),
        name="pull_request_conflicts",
    ),
    path(
        "pull-request/<str:pk>/reviews",
        views.PullRequestReviewListView.as_view(),
        name="pull_request_reviews",
    ),
    path(
        "pull-request/<str:pull_request>/reviews/add/",
        views.PullRequestReviewEditView.as_view(),
        name="pull_request_add_review",
    ),
    path(
        "pull-request/<str:pk>/commits",
        views.PullRequestCommitListView.as_view(),
        name="pull_request_commits",
    ),
]



"""utils.py contains utility methods used throughout the plugin."""


from contextlib import contextmanager
from copy import deepcopy

from django.db import connection, connections

from nautobot_version_control.constants import DB_NAME, DOLT_BRANCH_KEYWORD


class DoltError(Exception):
    """DoltError is a type of error to represent errors from the Dolt database custom functions."""

    pass  # pylint: disable=W0107


def author_from_user(user):
    """author_from_user returns an author string from a user object.

    Note that while user.email is optional in Django, it's mandatory for Dolt.
    """
    if user:
        if user.email:
            return f"{user.username} <{user.email}>"
        # RFC 6761 defines .invalid as a reserved TLD that will never be used for real-world domains
        return f"{user.username} <{user.username}@nautobot.invalid>"
    # default to generic user
    return "unknown <unknown@nautobot.invalid>"


def is_dolt_model(model):
    """
    Returns `True` if `instance` is an instance of
    a model from the Dolt plugin. Generally,
    """
    app_label = model._meta.app_label
    return app_label == "nautobot_version_control"


def alter_session_branch(sess=None, branch=None):
    """alter_session_branch sets the session cook branch to a new branch."""
    if sess is None or branch is None:
        raise ValueError("invalid args to change_branches()")
    sess[DOLT_BRANCH_KEYWORD] = branch


def active_branch():
    """active_branch returns the current active_branch from dolt."""
    with connection.cursor() as cursor:
        cursor.execute("SELECT active_branch() FROM dual;")
        return cursor.fetchone()[0]


def db_for_commit(commit):
    """
    Uses "database-revision" syntax
    adds a database entry for the commit
    e.g. "nautobot/3a5mqdgao8029bf8ji0huobbskq1n1l5"
    """
    cm_hash = str(commit)
    if len(cm_hash) != 32:
        raise Exception("commit hash length is incorrect")
    db = deepcopy(connections.databases["default"])
    db["id"] = cm_hash
    db["NAME"] = f"{DB_NAME}/{cm_hash}"
    connections.databases[cm_hash] = db
    return cm_hash


@contextmanager
def query_on_branch(branch):
    """query_on_branch checkout to another branch, runs a query, and checkouts back to main."""
    # TODO: remove in favor of db_for_commit
    with connection.cursor() as cursor:
        prev = active_branch()
        cursor.execute(f"""SELECT dolt_checkout("{branch}") FROM dual;""")
        yield
        cursor.execute(f"""SELECT dolt_checkout("{prev}") FROM dual;""")



"""Constants.py defines several important constants used throughout the plugin."""

# TODO: move these to settings?


DB_NAME = "nautobot"

GLOBAL_DB = "global"

DOLT_DEFAULT_BRANCH = "main"

DOLT_BRANCH_KEYWORD = "dolt-branch"



"""serializers.py implements serializers for different modules."""
from nautobot.core.api import ValidatedModelSerializer
from nautobot_version_control.models import Branch, Commit, PullRequest, PullRequestReview


class BranchSerializer(ValidatedModelSerializer):
    """BranchSerializer serializes a Branch"""

    class Meta:
        model = Branch
        fields = "__all__"


class CommitSerializer(ValidatedModelSerializer):
    """CommitSerializer serializes a Commit"""

    class Meta:
        model = Commit
        fields = "__all__"


class PullRequestSerializer(ValidatedModelSerializer):
    """PullRequestSerializer serializes a PullRequest"""

    class Meta:
        model = PullRequest
        fields = "__all__"


class PullRequestReviewSerializer(ValidatedModelSerializer):
    """PullRequestReviewSerializer serializes a PullRequestReview"""

    class Meta:
        model = PullRequestReview
        fields = "__all__"



# Generated by Django 3.1.12 on 2021-06-24 19:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("nautobot_version_control", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="BranchMeta",
            fields=[
                ("branch", models.TextField(primary_key=True, serialize=False)),
                ("source_branch", models.TextField()),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "author",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "nautobot_version_control_branchmeta",
            },
        ),
    ]



"""diff_factory wraps a model's diff and returns a queryable DiffModel."""

import copy

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils import timezone
from django.utils.html import format_html

import django_tables2 as tables
from django_tables2.utils import call_with_appropriate

from nautobot_version_control import diff_table_for_model


class DiffListViewFactory:
    """DiffListViewFactory dynamically generate diff models."""

    def __init__(self, content_type):
        self.ct = content_type

    def get_table_model(self):
        """get_table_model returns the underlying the underlying model."""
        try:
            return apps.get_model("nautobot_version_control", self.table_model_name)
        except LookupError:
            return self.make_table_model()

    def make_table_model(self):
        """make_table_model creates a DiffList of a model."""
        try:
            # lookup the list view table for this content type
            # todo: once available, use https://github.com/nautobot/nautobot/issues/747
            model = self.ct.model_class()
            ModelViewTable = diff_table_for_model(model)  # pylint: disable=C0103

            return type(
                self.table_model_name,
                (
                    ModelViewTable,
                    DiffListViewBase,
                ),
                {
                    "__module__": "nautobot_version_control.tables",
                    "_declared": timezone.now(),
                    "Meta": self._get_table_meta(ModelViewTable),
                    "content_type": self.ct,
                },
            )
        except KeyError as e:
            raise e

    def _get_table_meta(self, table):
        meta = copy.deepcopy(table._meta)
        # add diff styling
        meta.row_attrs = {"class": row_attrs_for_record}
        meta.sequence = ("diff", "...")
        return meta

    @property
    def table_model_name(self):
        """return the diff table for the model."""
        return f"diff_{str(self.ct.app_label)}_{str(self.ct.model)}"


def row_attrs_for_record(record):  # pylint: disable=R1710
    """row_attrs_for_record returns button attributes per diff type."""
    if not record.diff:
        return ""
    if record.diff["diff_type"] == "added":
        return "bg-success"
    if record.diff["diff_type"] == "removed":
        return "bg-danger"

    # diff_type == "modified"
    if record.diff["root"] == "to":
        return "bg-warning"
    if record.diff["root"] == "from":
        return "bg-warning"


class DiffListViewBase(tables.Table):
    """DiffListViewBase base model for a DiffList."""

    diff = tables.Column(verbose_name="Diff Type")

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for col in self.columns:
            if col.name == "diff":
                continue  # uses `render_diff()`
            col.render = self.wrap_render_func(col.render)

    def render_diff(self, value, record):  # pylint: disable=W0613
        """Custom rendering for the the `Diff Type` columns."""
        ct = ContentType.objects.get_for_model(self.Meta.model)  # pylint: disable=E1101
        href = reverse(
            "plugins:nautobot_version_control:diff_detail",
            kwargs={
                "app_label": ct.app_label,
                "model": ct.model,
                "from_commit": record.diff["from_commit"],
                "to_commit": record.diff["to_commit"],
                "pk": record.pk,
            },
        )

        if record.diff["diff_type"] == "added":
            return format_html(
                f"""<a href="{ href }">
                    <span class="label label-success">added</span>
                </a>"""
            )
        elif record.diff["diff_type"] == "removed":
            return format_html(
                f"""<a href="{ href }">
                    <span class="label label-danger">removed</span>
                </a>"""
            )
        else:  # diff_type == "modified"
            cnt = self.count_diffs(record.diff)
            return format_html(
                f"""<a href="{ href }">
                    <span class="label label-primary">changed ({ cnt })</span>
                </a>"""
            )

    @staticmethod
    def count_diffs(diff):
        """count_diffs counts the numbers of diffs."""
        skip_fields = (
            "root",
            "diff_type",
            "to_commit",
            "to_commit_date",
            "from_commit",
            "from_commit_date",
        )
        cnt = 0
        for k, v in diff.items():
            if k in skip_fields:
                continue
            if k.startswith("to_"):
                # compare to and from values
                from_key = f"from_{k[3:]}"
                if v != diff[from_key]:
                    cnt += 1
        return cnt

    @staticmethod
    def wrap_render_func(fn):
        """Wraps an existing cell rendering function with diff styling."""

        def render_before_after_diff(value, record, column, bound_column, bound_row, table):  # pylint: disable=R0913
            # the previous render function may take any of the
            # following args, so provide them all
            kwargs = {
                "value": value,
                "record": record,
                "column": column,
                "bound_column": bound_column,
                "bound_row": bound_row,
                "table": table,
            }
            try:
                # render the existing column function with best effort.
                cell = call_with_appropriate(fn, kwargs)
            except Exception:
                # In particular, rendering TemplateColumns for deleted rows
                # causes errors. Deleted rows are accessed with "time-travel"
                # queries, but are templates rendered from the current tip of
                # the branch, leading to referential integrity errors.
                return value

            if not record.diff or record.diff["diff_type"] != "modified":
                # only render before/after diff styling
                # for 'modified' rows
                return cell

            before_name = f"from_{bound_column.name}"
            if before_name not in record.diff:
                # can't render diff styling
                return cell

            after_name = f"to_{bound_column.name}"
            if after_name in record.diff and record.diff[after_name] == record.diff[before_name]:
                # no diff
                return cell

            # re-render the cell value with its before value
            kwargs["value"] = record.diff[before_name]
            before_cell = call_with_appropriate(fn, kwargs)

            if before_cell == cell:
                # no change
                return cell

            before_cell = before_cell if before_cell else " — "
            return format_html(
                f"""<div>
                <span class="bg-danger text-danger">
                    <b>{before_cell}</b>
                </span>
                </br>
                <span class="bg-success text-success">
                    <b>{cell}</b>
                </span>
            </div>"""
            )

        return render_before_after_diff



"""tests.py contains unittests for the nautobot version control plugin."""


from django.test import override_settings, TransactionTestCase
from django.urls import reverse
from django.db import connection

from nautobot.utilities.testing import APITestCase, APIViewTestCases
from nautobot.users.models import User
from nautobot.dcim.models import Manufacturer

from nautobot_version_control.models import Branch, Commit, PullRequest, PullRequestReview
from nautobot_version_control.merge import get_conflicts_count_for_merge
from nautobot_version_control.utils import active_branch
from nautobot_version_control.constants import DOLT_DEFAULT_BRANCH


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class DoltTestCase(TransactionTestCase):
    """DoltTestCase wraps test cases with the subsequent router setting."""

    databases = ["default", "global"]


class TestBranches(DoltTestCase):
    """TestBranch Tests the creation and deletion of branches."""

    default = DOLT_DEFAULT_BRANCH

    def setUp(self):
        """setUp is ran before every testcase."""
        self.user, _ = User.objects.get_or_create(
            username="branch-test", email="branch-test@example.com", is_superuser=True
        )
        self.user_no_email, _ = User.objects.get_or_create(username="branch-test-no-email", is_superuser=True)

    def tearDown(self):
        """tearDown is ran after every testcase."""
        Branch.objects.exclude(name=self.default).delete()

    def test_default_branch(self):
        """test_default_branch asserts that the main branch exists and is preinitialized as the active branch."""
        self.assertEqual(Branch.objects.filter(name=self.default).count(), 1)
        self.assertEqual(active_branch(), self.default)

    def test_create_branch(self):
        """test_create_branch tests the creation of a new branch."""
        Branch(name="another", starting_branch=self.default).save()
        self.assertEqual(Branch.objects.filter(name="another").count(), 1)

    def test_delete_with_pull_requests(self):
        """test_delete_with_pull_requests tests that deleting a branch cannot happen unless you delete a branch first."""
        Branch(name="todelete", starting_branch=self.default).save()
        PullRequest.objects.create(
            title="My Review",
            state=0,
            source_branch="todelete",
            destination_branch=self.default,
            description="review1",
            creator=self.user,
        )

        # Try to delete the branch
        try:
            Branch.objects.filter(name="todelete").delete()
            self.fail("the branch delete should've failed")
        except:  # nosec pylint: disable=W0702 # noqa
            pass

        # Delete the pr and try again
        PullRequest.objects.filter(title="My Review").delete()
        Branch.objects.filter(name="todelete").delete()
        self.assertEqual(Branch.objects.filter(name="todelete").count(), 0)

    def test_merge_ff(self):
        """test_merge_ff tests whether a ff merge works."""
        Branch(name="ff", starting_branch=self.default).save()
        main = Branch.objects.get(name=self.default)
        other = Branch.objects.get(name="ff")

        Commit(message="commit any changes").save(user=self.user)

        # Commit().save() doesn't update the PK correctly, it appears - .refresh_from_db() fails with
        # "Commit matching query does not exist". So we need to load it fresh from the DB instead:
        commit = Commit.objects.order_by("date").last()
        self.assertEqual(commit.committer, self.user.username)
        self.assertEqual(commit.email, self.user.email)

        # Checkout to the other branch and make a change
        other.checkout()
        Manufacturer.objects.all().delete()
        Manufacturer.objects.create(name="m1", slug="m-1")
        Commit(message="added a manufacturer").save(user=self.user_no_email)

        commit = Commit.objects.order_by("date").last()
        self.assertEqual(commit.committer, self.user_no_email.username)
        self.assertEqual(commit.email, f"{self.user_no_email.username}@nautobot.invalid")

        # Now do a merge
        main.checkout()
        main.merge(other, user=self.user)
        with connection.cursor() as cursor:
            cursor.execute("SELECT * FROM dolt_log where message=%s", ["added a manufacturer"])
            self.assertFalse(cursor.fetchone()[0] is None)

        # Verify the the main branch has the data
        self.assertEqual(Manufacturer.objects.filter(name="m1", slug="m-1").count(), 1)

    def test_merge_no_ff(self):
        """test_merge_no_ff tests whether a non-ff merge works."""
        Branch(name="noff", starting_branch=self.default).save()
        main = Branch.objects.get(name=self.default)
        other = Branch.objects.get(name="noff")

        # # Create a change on main
        Manufacturer.objects.create(name="m2", slug="m-2")
        Commit(message="commit m2").save(user=self.user)

        # Checkout to the other branch and make a change
        other.checkout()
        Manufacturer.objects.create(name="m3", slug="m-3")
        Commit(message="commit m3").save(user=self.user_no_email)

        # Now do a merge
        main.checkout()
        main.merge(other, user=self.user)
        with connection.cursor() as cursor:
            cursor.execute("SELECT * FROM dolt_log where message=%s", ["commit m3"])
            self.assertFalse(cursor.fetchone()[0] is None)

        # Verify the the main branch has the data
        self.assertEqual(Manufacturer.objects.filter(name="m2", slug="m-2").count(), 1)
        self.assertEqual(Manufacturer.objects.filter(name="m3", slug="m-3").count(), 1)

    def test_merge_conflicts(self):
        """test_merge_conflicts tests whether a merge with conflicts is detected and errors."""
        main = Branch.objects.get(name=self.default)
        Manufacturer.objects.all().delete()
        Manufacturer.objects.create(name="m2", slug="m-2")
        Commit(message="commit m2 with slug m-2").save(user=self.user)

        Branch(name="conflicts", starting_branch=self.default).save()
        other = Branch.objects.get(name="conflicts")

        # # Create a change on main
        Manufacturer.objects.filter(name="m2", slug="m-2").update(slug="m-15")
        Commit(message="commit m2 with slug m-15").save(user=self.user)

        # Checkout to the other branch and make a change
        other.checkout()
        Manufacturer.objects.filter(name="m2", slug="m-2").update(slug="m-16")
        Commit(message="commit m2 with slug m-16").save(user=self.user)

        # Now do a merge. It should throw an exception due to conflicts
        main.checkout()
        try:
            main.merge(other, user=self.user)
            self.fail("this should error for conflicts")
        except:  # nosec # pylint: disable=W0702 # noqa
            pass

        # Validate that any potential conflicts actually exist
        self.assertEqual(get_conflicts_count_for_merge(other, main), 1)
        main.checkout()  # need this because of post truncate action with TransactionTests


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class TestApp(APITestCase):
    """TestApp tests the availability of the root api endpoint."""

    databases = ["default", "global"]

    def test_root(self):
        """test_root tests getting the api root."""
        url = reverse("plugins-api:nautobot_version_control-api:api-root")
        response = self.client.get(f"{url}?format=api", **self.header)

        self.assertEqual(response.status_code, 200)


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class TestBranchesApi(APITestCase, APIViewTestCases):
    """TestBranchesApi tests the get,create,delete,etc. of the Branches api."""

    databases = ["default", "global"]
    model = Branch
    brief_fields = ["name", "starting_branch"]
    create_data = [
        {"name": "b4", "starting_branch": "b1"},
        {"name": "b5", "starting_branch": "b2"},
        {"name": "b6", "starting_branch": "b3"},
    ]

    @classmethod
    def setUpTestData(cls):
        """creates some test data for test suite."""
        Branch.objects.create(name="b1", starting_branch=DOLT_DEFAULT_BRANCH)
        Branch.objects.create(name="b2", starting_branch=DOLT_DEFAULT_BRANCH)
        Branch.objects.create(name="b3", starting_branch=DOLT_DEFAULT_BRANCH)

    # The ApiViewTestCase mixin handles get, create, etc. thoroughly but it's a useful exercise for readers to understand
    # what is happening in the background
    def test_get(self):
        """test_get tets the get method of the API."""
        url = reverse("plugins-api:nautobot_version_control-api:branch-list")
        self.add_permissions(f"{self.model._meta.app_label}.view_{self.model._meta.model_name}")
        response = self.client.get(f"{url}?format=json", **self.header)

        self.assertEqual(response.status_code, 200)

        data = response.json()
        self.assertTrue(data["count"] > 0)


class TestPullRequestApi(APITestCase, APIViewTestCases):
    """TestPullRequestApi tests the PullRequest api."""

    databases = ["default", "global"]
    model = PullRequest
    brief_fields = [
        "title",
        "state",
        "source_branch",
        "destination_branch",
        "description",
        "creator",
        "created_at",
    ]

    @classmethod
    def setUpTestData(cls):
        """setUpTestData setups test data."""
        User.objects.get_or_create(username="pr-reviewer", is_superuser=True)
        PullRequest.objects.create(
            title="Review 1",
            state=0,
            source_branch="b1",
            destination_branch="b2",
            description="review1",
            creator=User.objects.get(username="pr-reviewer"),
        )
        PullRequest.objects.create(
            title="Review 2",
            state=0,
            source_branch="b2",
            destination_branch="b3",
            description="review 2",
            creator=User.objects.get(username="pr-reviewer"),
        )
        PullRequest.objects.create(
            title="Review 3",
            state=0,
            source_branch="b3",
            destination_branch="b1",
            description="review 3",
            creator=User.objects.get(username="pr-reviewer"),
        )

    def test_get(self):
        """test_get tests the pull-request get api."""
        url = reverse("plugins-api:nautobot_version_control-api:pullrequest-list")
        self.add_permissions(f"{self.model._meta.app_label}.view_{self.model._meta.model_name}")
        response = self.client.get(f"{url}?format=json", **self.header)

        self.assertEqual(response.status_code, 200)

        data = response.json()
        self.assertEqual(data["count"], 3)


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class TestPullRequests(DoltTestCase):
    """TestPullRequests tests the functionality of the PullRequest model."""

    default = DOLT_DEFAULT_BRANCH

    def setUp(self):
        """setUp runs before every test case."""
        self.user = User.objects.get_or_create(username="branch-test", is_superuser=True)[0]
        self.main = Branch.objects.get(name=self.default)

    def tearDown(self):
        """tearDown runs after every test case."""
        PullRequest.objects.all().delete()
        Branch.objects.exclude(name=self.default).delete()

    def test_pull_requests_write_to_main(self):
        """test_pull_requests_write_to_main asserts that pull request objects hit the global db."""
        Branch(name="test", starting_branch=self.default).save()
        Branch(name="test2", starting_branch=self.default).save()
        test_branch = Branch.objects.get(name="test")

        # Create a PR from the test branch
        test_branch.checkout()
        PullRequest.objects.create(
            title="MyPr",
            state=0,
            source_branch=test_branch.name,
            destination_branch=self.default,
            description="Can I get a review",
            creator=self.user,
        )
        Commit(message="commit any changes").save(user=self.user)

        # Checkout to the main branch and verify that the pull request exists
        self.main.checkout()
        self.assertEqual(
            1,
            PullRequest.objects.filter(source_branch=test_branch.name, destination_branch=self.default).count(),
        )


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class TestPullRequestReviewsApi(APITestCase, APIViewTestCases):
    """TestPullRequestReviewsApi tests whether the PullRequestReview model api."""

    databases = ["default", "global"]
    model = PullRequestReview
    brief_fields = ["pull_request", "reviewer", "reviewed_at", "state", "summary"]

    @classmethod
    def setUpTestData(cls):
        """setupTestData create test data for the suite."""
        # Create the Pull Request Data before the reviews
        TestPullRequestApi.setUpTestData()
        PullRequestReview.objects.create(
            pull_request=PullRequest.objects.get(title="Review 1"),
            reviewer=User.objects.get(username="pr-reviewer"),
            state=1,
            summary="nice job",
        )
        PullRequestReview.objects.create(
            pull_request=PullRequest.objects.get(title="Review 2"),
            reviewer=User.objects.get(username="pr-reviewer"),
            state=0,
            summary="This is a comment",
        )

    def test_get(self):
        """test_get tests the PullRequestReview get_api."""
        url = reverse("plugins-api:nautobot_version_control-api:pullrequestreview-list")
        self.add_permissions(f"{self.model._meta.app_label}.view_{self.model._meta.model_name}")
        response = self.client.get(f"{url}?format=json", **self.header)

        self.assertEqual(response.status_code, 200)

        data = response.json()
        self.assertEqual(data["count"], 2)


@override_settings(DATABASE_ROUTERS=["nautobot_version_control.routers.GlobalStateRouter"])
class TestCommitsApi(APITestCase, APIViewTestCases):
    """TestCommitsApi tests the Commit model Api."""

    databases = ["default", "global"]
    model = Commit

    def test_root(self):
        """test commits."""
        url = reverse("plugins-api:nautobot_version_control-api:commit-list")
        self.add_permissions(f"{self.model._meta.app_label}.view_{self.model._meta.model_name}")
        response = self.client.get(f"{url}?format=api", **self.header)

        self.assertEqual(response.status_code, 200)



"""Injection of branch information banner via the Plugins API."""

from django.utils.html import format_html

from nautobot.extras.choices import BannerClassChoices
from nautobot.extras.plugins import PluginBanner

from nautobot_version_control.constants import DOLT_BRANCH_KEYWORD
from nautobot_version_control.utils import active_branch


def banner(context, *args, **kwargs):
    """Show a banner indicating the current active branch, if the user is logged in."""
    if not context.request.user.is_authenticated:
        return None
    branch_name = active_branch()
    return PluginBanner(
        content=format_html(
            """
<div class="text-center">
    Active Branch: <strong>{}</strong>
    <div class = "pull-right">
        <div class="btn btn-xs btn-primary" id="branch-share-button">
            Share
        </div>
    </div>
</div>
<script>
    const btn = document.getElementById("branch-share-button");
    btn.addEventListener('click', ()=>{{
        const currLink = window.location.href;
        const copiedLink = currLink + "?{}={}";
        navigator.clipboard.writeText(copiedLink);
        btn.textContent = "Copied!"
    }});
</script>""",
            branch_name,
            DOLT_BRANCH_KEYWORD,
            branch_name,
        ),
        banner_class=BannerClassChoices.CLASS_INFO,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nautobot_version_control',
 'nautobot_version_control.api',
 'nautobot_version_control.dynamic',
 'nautobot_version_control.graphql',
 'nautobot_version_control.management.commands',
 'nautobot_version_control.migrations']

package_data = \
{'': ['*'],
 'nautobot_version_control': ['templates/nautobot_version_control/*',
                              'templates/nautobot_version_control/pull_request/*']}

install_requires = \
['mysqlclient>=2.0.3,<3.0.0', 'nautobot>=1.2.0,<1.3.0']

setup_kwargs = {
    'name': 'nautobot-version-control',
    'version': '1.0.0a0',
    'description': 'Nautobot Version Control',
    'long_description': '# Nautobot Version Control\n\nThe Nautobot Version Control app brings version control to the [Nautobot](https://github.com/nautobot/nautobot) open source Network Source of Truth and Network Automation Platform. \n\nNautobot provides a number of features to validate its data model and safeguard network configuration from errors. Adding database versioning provides another layer of assurance by enabling human review of proposed changes to production data, use of automated testing pipelines, and database rollback in the case of errors. \n\nThe database versioning is made possible by the use of a [Dolt](https://github.com/dolthub/dolt) database. Dolt is a MySQL-compatible SQL database that you can fork, clone, branch, merge, push and pull just like a Git repository.\n\nDolt’s *branch* and *merge* versioning model allows operators to safely modify the data model on feature branches, merging to production only after validation is complete.\n\n## Documentation\n\nIn addition to this `README` file, there are docs covering the following topics:\n\n* [Version control operations](docs/version-control-operations.md)\n  * Covers common version control operations in the Version Control app\n* [Common workflows](docs/workflows/common_workflows.md)\n  * Covers common workflows enabled by the Version Control app\n* [Design](docs/design.md)\n  * Describes the technical design and implementation details of the Version Control app\n\n## Installation\n\n### Installation Considerations\n\nThere are some special considerations for running the Version Control app:\n\n* Nautobot 1.2.0 or later is required\n* The Nautobot installation **must** be running a Dolt database\n* There are some [additional configurations](#configuring-nautobot-to-use-version-control) required in `nautobot_config.py`\n\nThe version control app can be installed with pip3:\n\n```no-highlight\npip3 install nautobot-version-control\n```\n\nTo ensure the version control app is automatically reinstalled during future upgrades, create a new file named `local_requirements.txt` (if not already existing) in the Nautobot root directory (e.g. `/opt/nautobot`) to include the `nautobot-version-control` package:\n\n```no-highlight\necho nautobot-version-control >> local_requirements.txt\n```\n\n### Configuring Nautobot to use Version Control\n\nAdd this to your `nautobot_config.py` to prepare your Nautobot settings for Dolt:\n\n```python\n# Dolt requires a second database using the same credentials as the default database so that it may \n# generate diffs.\nDATABASES["global"] = DATABASES["default"]\n\n# Dolt requires a custom database router to generate the before & after queries for generating diffs.\nDATABASE_ROUTERS = ["dolt.routers.GlobalStateRouter"]\n\n# Because Dolt creates branches of the database, the default database sessions cannot be used. We \n# must tell Nautobot to use Redis for sessions instead. This adds a distinct cache configuration for\n# using Redis cache for sessions.\n# See: https://github.com/jazzband/django-redis#configure-as-session-backend\nCACHES["sessions"] = {\n    "BACKEND": "django_redis.cache.RedisCache",\n    "LOCATION": parse_redis_connection(redis_database=2),\n    "TIMEOUT": 300,\n    "OPTIONS": {\n        "CLIENT_CLASS": "django_redis.client.DefaultClient",\n    },\n}\n\n# Use the sessions alias defined in CACHES for sessions caching\nSESSION_CACHE_ALIAS = "sessions"\n\n# Use the Redis cache as the session engine\nSESSION_ENGINE = "django.contrib.sessions.backends.cache"\n\n# Enable the Version Control app\nPLUGINS = [ "nautobot_version_control" ]\n```\n\nThen run database migrations:\n\n```no-highlight\n$ nautobot-server migrate\n```\n\n### Production Deployment\n\nTBD as we progress towards stable release.\n\n## Development Environment\n\nIn addition to the instructions below, we also have a [demonstration video](https://youtu.be/XHrTHwhbZLc) that walks the user through the dev and test setup process described below. \n\n### Getting Started\n\nGetting started with development for the Version Control plugin is straightforward. It is modeled directly after the development environment of Nautobot itself, and should feel very familiar to anyone with Django development experience.\n\nThe Version Control app uses a Docker Compose environment to make it simple to manage dependencies like Dolt and Redis.\n\n### Install Invoke\n\nBecause it is used to execute all common Docker workflow tasks, Invoke must be installed for your user environment. \nOn most systems, if you\'re installing without root/superuser permissions, the default will install into your local user environment.\n\n```no-highlight\n$ pip3 install invoke\n```\n\nIf you run into issues, you may also deliberately tell pip3 to install into your user environment by adding the --user flag:\n\n```no-highlight\n$ pip3 install --user invoke\n```\n\n### List Invoke Tasks\n\nNow that you have an invoke command, list the tasks defined in tasks.py with `invoke --list`:\n\n```no-highlight\n$ invoke --list\nAvailable tasks:\n\n  bandit              Run bandit to validate basic static code security analysis.\n  black               Check Python code style with Black.\n  build               Build Nautobot docker image.\n  check-migrations    Check for missing migrations.\n  cli                 Launch a bash shell inside the running Nautobot container.\n  createsuperuser     Create a new Nautobot superuser account (default: "admin"), will prompt for password.\n  debug               Start Nautobot and its dependencies in debug mode.\n  destroy             Destroy all containers and volumes.\n  flake8              Check for PEP8 compliance and other style issues.\n  generate-packages   Generate all Python packages inside docker and copy the file locally under dist/.\n  hadolint            Check Dockerfile for hadolint compliance and other style issues.\n  load-data           Load data.\n  makemigrations      Perform makemigrations operation in Django.\n  migrate             Perform migrate operation in Django.\n  nbshell             Launch an interactive nbshell session.\n  post-upgrade        Performs Nautobot common post-upgrade operations using a single entrypoint.\n  pydocstyle          Run pydocstyle to validate docstring formatting adheres to NTC defined standards.\n  pylint              Run pylint code analysis.\n  restart             Gracefully restart all containers.\n  sphinx              Rebuild Sphinx documentation on changes, with live-reload in the\n                      browser.\n  start               Start Nautobot and its dependencies in detached mode.\n  stop                Stop Nautobot and its dependencies.\n  tests               Run all tests for this plugin.\n  unittest            Run Nautobot unit tests.\n  unittest-coverage   Report on code test coverage as measured by \'invoke unittest\'.\n  vscode              Launch Visual Studio Code with the appropriate Environment variables to run in a container.\n  yamllint            Run yamllint to validate formating adheres to NTC defined YAML standards.\n```\n\n### Using Docker with Invoke\n\nA development environment can be easily started up from the root of the project using commands detailed below.\n\n### Initialize the Local environment\n\nRun the following commmands to initialize the local environment:\n\n```\ncp development/creds.example.env development/creds.env\ninvoke build\n```\n\nFrom here, you can either [start the local environment with a sample database](#start-the-local-environment-with-a-sample-database) or [start the local environment with an empty database](#start-the-local-environment-with-an-empty-database).\n\n### Start the Local environment with a sample database\n\nThis is a good option for those who want to quickly spin up a working instance of Nautobot running the Version Control app. The steps below will create a Nautobot instance running Version Control and install sample data to experiment with.\n\n```\ninvoke migrate\ninvoke load-data\ninvoke start\n```\n\n> `invoke load-data` can take up to 30 min to run and it will generate many warning messages which may be safely ignored.\n\nAfter a few minutes, Nautobot will be available at `http://0.0.0.0:8080` \n\nYou can connect with either of these 2 accounts:\n\n* Login `admin` / Password `admin`\n* Login `demo` / Password `nautobot`\n\n\nRun the following commands to reset the local environment and load the sample dataset again:\n\n```\ninvoke stop\ninvoke destroy\ninvoke migrate\ninvoke load-data\ninvoke start\n```\n\n### Start the Local environment with an empty database\n\nThis option will simply start the local dev environment. Nautobot will have an empty database:\n\n```\ninvoke migrate\ninvoke start\n```\n\nAfter a few minutes, Nautobot will be available at `http://0.0.0.0:8080` \n\nYou can connect with:\n\n- Login `admin` / Password `admin`\n\nRun the following commands to reset the local environment:\n\n```\ninvoke stop\ninvoke destroy\ninvoke migrate\ninvoke start\n```\n',
    'author': 'Network to Code, LLC',
    'author_email': 'opensource@networktocode.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nautobot/nautobot-plugin-version-control',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



"""tables.py contains modules for implementing view tables for different models."""

import django_tables2 as tables
from django_tables2 import A
from nautobot.utilities.tables import BaseTable, ToggleColumn, ButtonsColumn

from nautobot_version_control.models import (
    Branch,
    Conflicts,
    Commit,
    PullRequest,
)

__all__ = ("BranchTable", "ConflictsSummaryTable", "CommitTable", "PullRequestTable")


#
# Branches
#

BRANCH_TABLE_BADGES = """
<div>
   {% if not record.active %}
        <a href="{% url 'plugins:nautobot_version_control:branch_checkout' pk=record.pk %}" class="btn btn-xs btn-primary" title="activate">
            Activate
        </a>
   {% endif %}
    <a href="{% url 'plugins:nautobot_version_control:pull_request_add' %}?source_branch={{ record.pk }}" class="btn btn-xs btn-success" title="pull_request">
        Pull Request
    </a>
    <a href="{% url 'plugins:nautobot_version_control:pull_request_add' %}?source_branch={{ default_branch }}&destination_branch={{ record.pk }}" class="btn btn-xs btn-warning" title="catch_up">
        Catchup
    </a>
</div>
"""


ACTIVE_BRANCH_BADGE = """
{% if record.active %}
    <div class="btn btn-xs btn-success" title="active">
        Active
    </div>
{% endif %}
"""


class BranchTable(BaseTable):
    """BranchTable render the table in the branch_list view."""

    pk = ToggleColumn()
    name = tables.LinkColumn()
    status = tables.TemplateColumn(ACTIVE_BRANCH_BADGE)
    hash = tables.LinkColumn("plugins:nautobot_version_control:commit", args=[A("hash")])
    actions = ButtonsColumn(
        Branch,
        pk_field="name",
        buttons=("checkout",),
        prepend_template=BRANCH_TABLE_BADGES,
    )
    ahead_behind = tables.Column(accessor=A("ahead_behind"), verbose_name="Ahead / Behind")
    starting_branch = tables.Column(accessor=A("source_branch"), verbose_name="Starting Branch")

    class Meta(BaseTable.Meta):
        model = Branch
        fields = (
            "pk",
            "name",
            "hash",
            "status",
            "ahead_behind",
            "created_by",
            "latest_committer",
            "latest_commit_date",
            "latest_commit_message",
            "starting_branch",
            "actions",
        )
        default_columns = (
            "name",
            "status",
            "ahead_behind",
            "created_by",
            "latest_committer",
            "latest_commit_date",
            "starting_branch",
            "actions",
        )


#
# Commits
#


class CommitTable(BaseTable):
    """CommitTable renders the commit table in the commit list view."""

    pk = ToggleColumn(visible=True)
    short_message = tables.LinkColumn(verbose_name="Commit Message")

    class Meta(BaseTable.Meta):
        model = Commit
        fields = (
            "pk",
            "short_message",
            "date",
            "committer",
            "email",
            "commit_hash",
        )
        default_columns = fields


class CommitRevertTable(BaseTable):
    """CommitRevert renders the commit revert table for the commit review view."""

    short_message = tables.LinkColumn(verbose_name="Commit Message")

    class Meta(BaseTable.Meta):
        model = Commit
        fields = (
            "short_message",
            "date",
            "committer",
            "email",
            "commit_hash",
        )
        default_columns = fields


#
# Conflicts
#


class ConflictsSummaryTable(BaseTable):
    """Summary table for `Conflicts` and `ConstraintViolations`."""

    class Meta(BaseTable.Meta):
        model = Conflicts
        fields = ("table", "num_conflicts", "num_violations")
        default_columns = fields


CONFLICT_TABLE_JSON = """
{% load helpers %}
<div class="rendered-json-data">
    <pre>{{ record.conflicts|render_yaml }}</pre>
</div>
"""


class ConflictsTable(BaseTable):
    """ConflictsTable renders the table in the conflict list view."""

    conflicts = tables.TemplateColumn(
        template_code=CONFLICT_TABLE_JSON,
        verbose_name="Conflicts",
    )

    class Meta(BaseTable.Meta):
        model = Conflicts
        fields = (
            "model",
            "id",
            "conflicts",
        )
        default_columns = fields


class ConstraintViolationsTable(BaseTable):
    """ConstraintViolations renders the table in the constraint violations table."""

    class Meta(BaseTable.Meta):
        model = Conflicts
        fields = (
            "model",
            "id",
            "violations",
        )
        default_columns = fields


#
# PullRequest
#

PR_STATUS_BADGES = """
<div>
{% if record.status == "open" %}
    <span class="label label-primary">Open</span>
{% elif record.status == "in-review" %}
    <span class="label label-primary">In Review</span>
{% elif record.status == "blocked" %}
    <span class="label label-warning">Blocked</span>
{% elif record.status == "approved" %}
    <span class="label label-success">Approved</span>
{% elif record.status == "merged" %}
    <span class="label label-info">Merged</span>
{% elif record.status == "closed" %}
    <span class="label label-danger">Closed</span>
{% else %}
    <span> - </span>
{% endif %}
</div>
"""


class PullRequestTable(BaseTable):
    """PullRequestTable renders the table in the pull request view."""

    pk = ToggleColumn()
    status = tables.TemplateColumn(
        template_code=PR_STATUS_BADGES,
        verbose_name="Status",
    )
    title = tables.LinkColumn()

    class Meta(BaseTable.Meta):
        model = PullRequest
        fields = (
            "pk",
            "title",
            "status",
            "source_branch",
            "destination_branch",
            "creator",
            "created_at",
        )
        default_columns = fields



"""urls.py implements api urls."""

from nautobot.core.api import OrderedDefaultRouter
from . import views


router = OrderedDefaultRouter()
router.APIRootView = views.VCSRootView

# Sites
router.register("branches", views.BranchViewSet)
router.register("commits", views.CommitViewSet)
router.register("pull_requests", views.PullRequestViewSet)
router.register("pull_requests_reviews", views.PullRequestReviewViewSet)

app_name = "nautobot_version_control-api"
urlpatterns = router.urls
